from typing import Optional, List, Dict, Any, Callable
from types import SimpleNamespace
from enum import Enum
from pathlib import Path
import os
import time
import shutil
import subprocess
from concurrent.futures import Future

from alipcs_py.alipcs import AliPCSApi, PcsFile
from alipcs_py.alipcs.pcs import PCS_UA
from alipcs_py.utils import human_size_to_int
from alipcs_py.common import constant
from alipcs_py.common.io import to_decryptio, DecryptIO, READ_SIZE
from alipcs_py.common.downloader import MeDownloader
from alipcs_py.common.progress_bar import (
    _progress,
    init_progress_bar,
    remove_progress_task,
    reset_progress_task,
)
from alipcs_py.commands.sifter import Sifter, sift
from alipcs_py.commands.log import get_logger

_print = print

from rich import print
from rich.progress import TaskID

logger = get_logger(__name__)

USER_AGENT = PCS_UA

DEFAULT_CONCURRENCY = 5
DEFAULT_CHUNK_SIZE = str(1 * constant.OneM)

# This is the threshold of range request setted by Ali server
MAX_CHUNK_SIZE = 50 * constant.OneM


class DownloadParams(SimpleNamespace):
    concurrency: int = DEFAULT_CONCURRENCY
    chunk_size: str = DEFAULT_CHUNK_SIZE
    quiet: bool = False


DEFAULT_DOWNLOADPARAMS = DownloadParams()


class Downloader(Enum):
    me = "me"
    aget_py = "aget"  # https://github.com/PeterDing/aget
    aget_rs = "ag"  # https://github.com/PeterDing/aget-rs
    aria2 = "aria2c"  # https://github.com/aria2/aria2

    # No use axel. It Can't handle URLs of length over 1024
    # axel = 'axel'  # https://github.com/axel-download-accelerator/axel

    # No use wget. the file url of alipan only supports `Range` request

    def which(self) -> Optional[str]:
        return shutil.which(self.value)

    def download(
        self,
        url: str,
        localpath: str,
        downloadparams: DownloadParams = DEFAULT_DOWNLOADPARAMS,
        out_cmd: bool = False,
        encrypt_password: bytes = b"",
    ):
        global DEFAULT_DOWNLOADER
        if not self.which():
            self = DEFAULT_DOWNLOADER

        localpath_tmp = localpath + ".tmp"

        def done_callback(fut: Future):
            err = fut.exception()
            if not err:
                shutil.move(localpath_tmp, localpath)
            else:
                logger.info("`download`: MeDownloader fails: error: %s", err)

        if self == Downloader.me:
            self._me_download(
                url,
                localpath_tmp,
                downloadparams=downloadparams,
                done_callback=done_callback,
                encrypt_password=encrypt_password,
            )
            return
        elif self == Downloader.aget_py:
            cmd = self._aget_py_cmd(url, localpath_tmp, downloadparams)
        elif self == Downloader.aget_rs:
            cmd = self._aget_rs_cmd(url, localpath_tmp, downloadparams)
        elif self == Downloader.aria2:
            cmd = self._aria2_cmd(url, localpath_tmp, downloadparams)
        else:
            cmd = self._aget_py_cmd(url, localpath_tmp, downloadparams)

        # Print out command
        if out_cmd:
            _print(" ".join((repr(c) for c in cmd)))
            return

        returncode = self.spawn(cmd, downloadparams.quiet)

        logger.debug("`download`: cmd returncode: %s", returncode)

        if returncode != 0:
            print(f"[italic]{self.value}[/italic] fails. return code: [red]{returncode}[/red]")
        else:
            if encrypt_password:
                dio = to_decryptio(open(localpath_tmp, "rb"), encrypt_password)
                if isinstance(dio, DecryptIO):
                    with open(localpath, "wb") as fd:
                        while True:
                            buf = dio.read(READ_SIZE)
                            if not buf:
                                break
                            fd.write(buf)

                    os.remove(localpath_tmp)
                    return
            shutil.move(localpath_tmp, localpath)

    def spawn(self, cmd: List[str], quiet: bool = False):
        child = subprocess.run(cmd, stdout=subprocess.DEVNULL if quiet else None)
        return child.returncode

    def _me_download(
        self,
        url: str,
        localpath: str,
        downloadparams: DownloadParams = DEFAULT_DOWNLOADPARAMS,
        done_callback: Optional[Callable[[Future], Any]] = None,
        encrypt_password: bytes = b"",
    ):
        headers = {
            "Referer": "https://www.aliyundrive.com/",
            "User-Agent": USER_AGENT,
            "Connection": "Keep-Alive",
        }

        task_id: Optional[TaskID] = None
        if not downloadparams.quiet:
            init_progress_bar()
            task_id = _progress.add_task("MeDownloader", start=False, title=localpath)

        def _wrap_done_callback(fut: Future):
            remove_progress_task(task_id)
            if done_callback:
                done_callback(fut)

        def monit_callback(task_id: Optional[TaskID], offset: int):
            if task_id is not None:
                _progress.update(task_id, completed=offset + 1)

        def except_callback(task_id: Optional[TaskID]):
            reset_progress_task(task_id)

        chunk_size_int = human_size_to_int(downloadparams.chunk_size)
        meDownloader = MeDownloader(
            "GET",
            url,
            headers=headers,
            max_workers=downloadparams.concurrency,
            max_chunk_size=chunk_size_int,
            callback=monit_callback,
            encrypt_password=encrypt_password,
        )

        if task_id is not None:
            length = len(meDownloader)
            _progress.update(task_id, total=length)
            _progress.start_task(task_id)

        meDownloader.download(
            Path(localpath),
            task_id=task_id,
            continue_=True,
            done_callback=_wrap_done_callback,
            except_callback=except_callback,
        )

    def _aget_py_cmd(
        self,
        url: str,
        localpath: str,
        downloadparams: DownloadParams = DEFAULT_DOWNLOADPARAMS,
    ):
        cmd = [
            self.which(),
            url,
            "-o",
            localpath,
            "-H",
            f"User-Agent: {USER_AGENT}",
            "-H",
            "Connection: Keep-Alive",
            "-H",
            "Referer: https://www.aliyundrive.com/",
            "-s",
            str(downloadparams.concurrency),
            "-k",
            downloadparams.chunk_size,
        ]
        return cmd

    def _aget_rs_cmd(
        self,
        url: str,
        localpath: str,
        downloadparams: DownloadParams = DEFAULT_DOWNLOADPARAMS,
    ):
        cmd = [
            self.which(),
            url,
            "-o",
            localpath,
            "-H",
            f"User-Agent: {USER_AGENT}",
            "-H",
            "Connection: Keep-Alive",
            "-H",
            "Referer: https://www.aliyundrive.com/",
            "-s",
            str(downloadparams.concurrency),
            "-k",
            downloadparams.chunk_size,
        ]
        return cmd

    def _aria2_cmd(
        self,
        url: str,
        localpath: str,
        downloadparams: DownloadParams = DEFAULT_DOWNLOADPARAMS,
    ):
        directory, filename = os.path.split(localpath)
        cmd = [
            self.which(),
            "-c",
            "--dir",
            directory,
            "-o",
            filename,
            "--header",
            f"User-Agent: {USER_AGENT}",
            "--header",
            "Connection: Keep-Alive",
            "--header",
            "Referer: https://www.aliyundrive.com/",
            "-s",
            str(downloadparams.concurrency),
            "-k",
            downloadparams.chunk_size,
            url,
        ]
        return cmd


DEFAULT_DOWNLOADER = Downloader.me


def download_file(
    api: AliPCSApi,
    pcs_file: PcsFile,
    localdir: str,
    share_id: str = None,
    downloader: Downloader = DEFAULT_DOWNLOADER,
    downloadparams: DownloadParams = DEFAULT_DOWNLOADPARAMS,
    out_cmd: bool = False,
    encrypt_password: bytes = b"",
):
    localpath = Path(localdir) / pcs_file.name

    # Make sure parent directory existed
    if not localpath.parent.exists():
        localpath.parent.mkdir(parents=True)

    if not out_cmd and localpath.exists():
        print(f"[yellow]{localpath}[/yellow] is ready existed.")
        return

    if not pcs_file:
        return

    if downloader != Downloader.me:
        print(f"[italic blue]Download[/italic blue]: {pcs_file.path or pcs_file.name} to {localpath}")

    download_url: Optional[str]
    if share_id:
        remote_temp_dir = "/__alipcs_py_temp__"
        pcs_temp_dir = api.path(remote_temp_dir) or api.makedir_path(remote_temp_dir)
        pcs_file = api.transfer_shared_files([pcs_file.file_id], pcs_temp_dir.file_id, share_id)[0]

        while True:
            pcs_file = api.meta(pcs_file.file_id)[0]
            if pcs_file.download_url:
                break
            time.sleep(2)

    if not pcs_file or pcs_file.is_dir:
        return

    pcs_file.update_download_url(api)
    download_url = pcs_file.download_url

    assert download_url

    downloader.download(
        download_url,
        str(localpath),
        downloadparams=downloadparams,
        out_cmd=out_cmd,
        encrypt_password=encrypt_password,
    )

    if share_id:
        api.remove(pcs_file.file_id)


def download_dir(
    api: AliPCSApi,
    pcs_file: PcsFile,
    localdir: str,
    share_id: str = None,
    sifters: List[Sifter] = [],
    recursive: bool = False,
    from_index: int = 0,
    downloader: Downloader = DEFAULT_DOWNLOADER,
    downloadparams=DEFAULT_DOWNLOADPARAMS,
    out_cmd: bool = False,
    encrypt_password: bytes = b"",
):
    remotefiles = list(api.list_iter(pcs_file.file_id, share_id=share_id))
    remotefiles = sift(remotefiles, sifters, recursive=recursive)
    for rp in remotefiles[from_index:]:
        if rp.is_file:
            download_file(
                api,
                rp,
                localdir,
                share_id=share_id,
                downloader=downloader,
                downloadparams=downloadparams,
                out_cmd=out_cmd,
                encrypt_password=encrypt_password,
            )
        else:  # is_dir
            if recursive:
                _localdir = Path(localdir) / os.path.basename(rp.path)
                download_dir(
                    api,
                    rp,
                    str(_localdir),
                    share_id=share_id,
                    sifters=sifters,
                    recursive=recursive,
                    from_index=from_index,
                    downloader=downloader,
                    downloadparams=downloadparams,
                    out_cmd=out_cmd,
                    encrypt_password=encrypt_password,
                )


def download(
    api: AliPCSApi,
    remotepaths: List[str],
    file_ids: List[str],
    localdir: str,
    share_id: str = None,
    sifters: List[Sifter] = [],
    recursive: bool = False,
    from_index: int = 0,
    downloader: Downloader = DEFAULT_DOWNLOADER,
    downloadparams: DownloadParams = DEFAULT_DOWNLOADPARAMS,
    out_cmd: bool = False,
    encrypt_password: bytes = b"",
):
    """Download `remotepaths` to the `localdir`

    Args:
        `from_index` (int): The start index of downloading entries from EACH remote directory
    """

    logger.debug(
        "`download`: sifters: %s, recursive: %s, from_index: %s, "
        "downloader: %s, downloadparams: %s, out_cmd: %s, has encrypt_password: %s",
        sifters,
        recursive,
        from_index,
        downloader,
        downloadparams,
        out_cmd,
        bool(encrypt_password),
    )

    for rp in remotepaths:
        rpf = api.path(rp, share_id=share_id)
        if not rpf:
            print(f"[yellow]WARNING[/yellow]: `{rp}` does not exist.")
            continue

        if rpf.is_file:
            download_file(
                api,
                rpf,
                localdir,
                share_id=share_id,
                downloader=downloader,
                downloadparams=downloadparams,
                out_cmd=out_cmd,
                encrypt_password=encrypt_password,
            )
        else:
            _localdir = str(Path(localdir) / rpf.name)
            download_dir(
                api,
                rpf,
                _localdir,
                share_id=share_id,
                sifters=sifters,
                recursive=recursive,
                from_index=from_index,
                downloader=downloader,
                downloadparams=downloadparams,
                out_cmd=out_cmd,
                encrypt_password=encrypt_password,
            )

    for file_id in file_ids:
        rpf = api.meta(file_id, share_id=share_id)[0]
        if not rpf:
            print(f"[yellow]WARNING[/yellow]: file_id `{file_id}` does not exist.")
            continue

        if rpf.is_file:
            download_file(
                api,
                rpf,
                localdir,
                share_id=share_id,
                downloader=downloader,
                downloadparams=downloadparams,
                out_cmd=out_cmd,
                encrypt_password=encrypt_password,
            )
        else:
            _localdir = str(Path(localdir) / rpf.name)
            download_dir(
                api,
                rpf,
                _localdir,
                share_id=share_id,
                sifters=sifters,
                recursive=recursive,
                from_index=from_index,
                downloader=downloader,
                downloadparams=downloadparams,
                out_cmd=out_cmd,
                encrypt_password=encrypt_password,
            )

    if downloader == Downloader.me:
        MeDownloader._exit_executor()

    _progress.stop()



from typing import Optional, List, Tuple
from pathlib import Path

from alipcs_py.alipcs import AliPCSApi, PcsFile, FromTo
from alipcs_py.common.path import walk, join_path
from alipcs_py.common.crypto import calc_sha1
from alipcs_py.common.constant import CPU_NUM
from alipcs_py.common.io import EncryptType
from alipcs_py.commands.upload import UploadType, upload as _upload, DEFAULT_SLICE_SIZE
from alipcs_py.commands.log import get_logger

from rich import print

logger = get_logger(__name__)


def sync(
    api: AliPCSApi,
    localdir: str,
    remotedir: str,
    encrypt_password: bytes = b"",
    encrypt_type: EncryptType = EncryptType.No,
    max_workers: int = CPU_NUM,
    slice_size: int = DEFAULT_SLICE_SIZE,
    show_progress: bool = True,
    rapiduploadinfo_file: Optional[str] = None,
    user_id: Optional[str] = None,
    user_name: Optional[str] = None,
):
    localdir = Path(localdir).as_posix()
    remotedir = Path(remotedir).as_posix()

    rpf = api.makedir_path(remotedir)
    assert rpf and rpf.is_dir, "remotedir must be a directory"

    if remotedir == "/":
        remotedir_len = 0
    else:
        remotedir_len = len(remotedir)

    all_pcs_files = {
        pcs_file.path[remotedir_len + 1 :]: pcs_file
        for pcs_file in api.list_path_iter(remotedir, recursive=True, include_dir=False)
    }

    fts: List[FromTo] = []
    check_list: List[Tuple[str, PcsFile]] = []
    all_localpaths = set()
    for localpath in walk(localdir):
        path = localpath[len(localdir) + 1 :]
        all_localpaths.add(path)

        if path not in all_pcs_files:
            fts.append(FromTo(localpath, join_path(remotedir, path)))
        else:
            check_list.append((localpath, all_pcs_files[path]))

    for lp, pf in check_list:
        sha1 = calc_sha1(Path(lp).open("rb"))

        if pf.rapid_upload_info and sha1.lower() != pf.rapid_upload_info.content_hash.lower():
            fts.append(FromTo(lp, pf.path))

    need_deleted_file_ids = []
    for rp in all_pcs_files.keys():
        if rp not in all_localpaths:
            need_deleted_file_ids.append(all_pcs_files[rp].file_id)

    logger.debug(
        "`sync`: all localpaths: %s, localpaths needed to upload: %s, remotepaths needed to delete: %s",
        len(all_localpaths),
        len(fts),
        len(need_deleted_file_ids),
    )

    _upload(
        api,
        fts,
        upload_type=UploadType.Many,
        check_name_mode="overwrite",
        encrypt_password=encrypt_password,
        encrypt_type=encrypt_type,
        max_workers=max_workers,
        slice_size=slice_size,
        show_progress=show_progress,
        rapiduploadinfo_file=rapiduploadinfo_file,
        user_id=user_id,
        user_name=user_name,
    )

    if need_deleted_file_ids:
        api.remove(*need_deleted_file_ids)
        print(f"Delete: [i]{len(need_deleted_file_ids)}[/i] remote paths")



from typing import Dict, List, Iterable, Optional
from collections import OrderedDict

from alipcs_py.alipcs.api import AliPCSApi
from alipcs_py.alipcs.inner import PcsFile
from alipcs_py.common.path import split_posix_path


class _Node:
    def __init__(self, file_id: str, pcs_file: PcsFile):
        self.file_id = file_id
        self.pcs_file = pcs_file
        self.sub_nodes: OrderedDict[str, _Node] = OrderedDict()


class PathTree:
    def __init__(self, api: AliPCSApi):
        self._api = api
        self.root = _Node("root", PcsFile.root())

    def search(self, remotepath: str = "", topdown: Iterable[str] = [], root: _Node = None) -> Optional[PcsFile]:
        if not topdown:
            assert remotepath.startswith("/")
            topdown = split_posix_path(remotepath)

        root = root or self.root
        return self._dfs(list(topdown), root)

    def _dfs(self, topdown: List[str], root: _Node) -> Optional[PcsFile]:
        if not topdown:
            return root.pcs_file

        next_key = topdown[0]
        root_pcs_file = root.pcs_file

        if next_key not in root.sub_nodes:
            for pf in self._api.list_iter(root_pcs_file.file_id):
                root.sub_nodes[pf.name] = _Node(pf.file_id, pf)
                if pf.name == next_key:
                    break

        if next_key not in root.sub_nodes:
            return None

        return self._dfs(topdown[1:], root.sub_nodes[next_key])



from typing import Optional, List, Dict
from dataclasses import dataclass
from pathlib import Path
import pickle

from alipcs_py.alipcs import AliPCSApi, AliPCSApiMix, PcsUser
from alipcs_py.common.path import PathType, join_path


@dataclass
class Account:
    user: PcsUser

    # Account Name which can be set by hand
    account_name: str = ""

    # current working directory
    pwd: str = "/"
    encrypt_password: Optional[str] = None

    # Is the account is used now
    on: bool = False

    def pcsapi(self) -> AliPCSApi:
        web_refresh_token = self.user.web_refresh_token
        web_access_token = self.user.web_access_token
        web_token_type = self.user.web_token_type or "Bearer"
        web_expire_time = self.user.web_expire_time or 0

        openapi_refresh_token = self.user.openapi_refresh_token
        openapi_access_token = self.user.openapi_access_token
        openapi_token_type = self.user.openapi_token_type or "Bearer"
        openapi_expire_time = self.user.openapi_expire_time or 0

        client_id = self.user.client_id
        client_secret = self.user.client_secret
        client_server = self.user.client_server

        user_id = self.user.user_id or ""
        user_name = self.user.user_name or ""
        nick_name = self.user.nick_name or ""
        device_id = self.user.device_id or ""
        default_drive_id = self.user.default_drive_id or ""
        role = self.user.role or ""
        status = self.user.status or ""

        assert web_refresh_token, f"{self}.user.web_refresh_token is None"
        return AliPCSApiMix(
            web_refresh_token,
            web_access_token=web_access_token or "",
            web_token_type=web_token_type,
            web_expire_time=web_expire_time,
            openapi_refresh_token=openapi_refresh_token or "",
            openapi_access_token=openapi_access_token or "",
            openapi_token_type=openapi_token_type,
            openapi_expire_time=openapi_expire_time,
            client_id=client_id or "",
            client_secret=client_secret or "",
            client_server=client_server or "",
            user_id=user_id,
            user_name=user_name,
            nick_name=nick_name,
            device_id=device_id,
            default_drive_id=default_drive_id,
            role=role,
            status=status,
        )

    @staticmethod
    def from_refresh_token(
        web_refresh_token: str,
        web_access_token: str = "",
        web_token_type: str = "Bearer",
        web_expire_time: int = 0,
        openapi_refresh_token: str = "",
        openapi_access_token: str = "",
        openapi_token_type: str = "Bearer",
        openapi_expire_time: int = 0,
        client_id: str = "",
        client_secret: str = "",
        client_server: str = "",
        account_name: str = "",
    ) -> "Account":
        api = AliPCSApiMix(
            web_refresh_token,
            web_access_token=web_access_token or "",
            web_token_type=web_token_type,
            web_expire_time=web_expire_time,
            openapi_refresh_token=openapi_refresh_token or "",
            openapi_access_token=openapi_access_token or "",
            openapi_token_type=openapi_token_type,
            openapi_expire_time=openapi_expire_time,
            client_id=client_id or "",
            client_secret=client_secret or "",
            client_server=client_server or "",
        )
        user = api.user_info()

        return Account(
            user,
            account_name or user.user_id,  # Default use `user_id`
        )


class AccountManager:
    """Account Manager

    Manage all accounts
    """

    def __init__(self, data_path: Optional[PathType] = None):
        self._accounts: Dict[str, Account] = {}  # user_id (str) -> Account
        self._who: Optional[str] = None  # user_id (str)
        self._data_path = data_path
        self._apis: Dict[str, AliPCSApi] = {}

    @staticmethod
    def load_data(data_path: PathType) -> "AccountManager":
        try:
            data_path = Path(data_path).expanduser()
            am = pickle.load(data_path.open("rb"))
            am._data_path = data_path
            return am
        except Exception:
            return AccountManager(data_path=data_path)

    @property
    def accounts(self) -> List[Account]:
        """All accounts"""

        return list(self._accounts.values())

    def get_api(self, user_id: Optional[str] = None) -> Optional[AliPCSApi]:
        user_id = user_id or self._who
        if user_id:
            account = self._accounts.get(user_id)
            if not account:
                return None

            if user_id in self._apis:
                api = self._apis[user_id]
            else:
                api = account.pcsapi()
                self._apis[user_id] = api
            return api
        return None

    def set_account_name(self, account_name: str, user_id: Optional[str] = None):
        """Set account name"""

        user_id = user_id or self._who

        assert user_id, "No recent user"

        account = self._accounts.get(user_id)

        assert account

        account.account_name = account_name

    def set_encrypt_password(self, encrypt_password: Optional[str] = None):
        """Set encryption key"""

        assert self._who, "No recent user"

        account = self._accounts.get(self._who)

        assert account

        account.encrypt_password = encrypt_password

    def cd(self, remotedir: str = "/"):
        """Change current working directory"""

        assert self._who, "No recent user"

        account = self._accounts.get(self._who)

        assert account

        pwd = join_path(account.pwd, remotedir)
        account.pwd = pwd

    @property
    def pwd(self) -> str:
        """Current working directory of recent user"""

        assert self._who, "No recent user"

        account = self._accounts.get(self._who)

        assert account

        return account.pwd

    def who(self, user_id: Optional[str] = None) -> Optional[Account]:
        """Return recent `Account`"""

        user_id = user_id or self._who
        if user_id:
            return self._accounts.get(user_id)
        else:
            return None

    def update(self, user_id: Optional[str] = None):
        """Update user_info"""

        user_id = user_id or self._who
        if user_id:
            api = self.get_api(user_id)
            if api:
                account = self._accounts.get(user_id)
                assert account

                user = api.user_info()
                account.user = user

    def su(self, user_id: str):
        """Change recent user with `PcsUser.user_id`

        Args:
            user_id (str): `PcsUser.user_id`
        """

        assert user_id in self._accounts, f"No user {user_id}"

        self._switch(user_id)

    def _switch(self, user_id: str):
        for uid, account in list(self._accounts.items()):
            if uid == user_id:
                account.on = True
                self._who = account.user.user_id
            else:
                account.on = False

    def add_account(self, account: Account):
        """Add an account to the manager"""

        self._accounts[account.user.user_id] = account

    def delete_account(self, user_id: str):
        """Delete an account

        Args:
            user_id (str): `PcsUser.user_id`
        """

        if user_id in self._accounts:
            del self._accounts[user_id]
        if user_id == self._who:
            self._who = None

    def save(self, data_path: Optional[PathType] = None):
        """Serialize to local path"""

        data_path = data_path or self._data_path
        assert data_path, "No data path"

        data_path = Path(data_path).expanduser()
        if not data_path.parent.exists():
            data_path.parent.mkdir(parents=True)

        apis = self._apis
        self._apis = {}  # Ignore to save apis

        pickle.dump(self, open(data_path, "wb"))

        self._apis = apis



from typing import TypeVar, Optional, List, Union, Pattern
from abc import ABC
import re

from alipcs_py.alipcs import PcsFile


class Sifter(ABC):
    def pattern(self) -> Union[Pattern, str, None]:
        """
        The regex pattern used to the sifter

        If it returns '', then the sifter will match all inputs
        """
        return None

    def include(self) -> bool:
        """Include the sifted result or exclude"""
        return True

    def sift(self, obj: Union[PcsFile, str]) -> bool:
        """
        True: to include
        False: to exclude
        """

        include = self.include()
        pat = self.pattern()

        if not pat:
            return include

        if isinstance(obj, PcsFile):
            buf = obj.path
        else:
            buf = obj

        if isinstance(pat, Pattern):
            if pat.search(buf):
                return include
            else:
                return not include
        else:  # str
            if pat in buf:
                return include
            else:
                return not include

    def __call__(self, obj: Union[PcsFile, str]) -> bool:
        return self.sift(obj)


class IncludeSifter(Sifter):
    def __init__(self, needle: Optional[str], regex: bool = False):
        _pattern: Union[Pattern, str, None] = None
        self._pattern = _pattern
        if needle:
            if regex:
                self._pattern = re.compile(needle)
            else:
                self._pattern = needle

    def pattern(self):
        return self._pattern


class ExcludeSifter(IncludeSifter):
    def __init__(self, needle: Optional[str], regex: bool = False):
        super().__init__(needle, regex=regex)

    def include(self):
        return False


class IsFileSifter(Sifter):
    def sift(self, obj: Union[PcsFile, str]) -> bool:
        assert isinstance(obj, PcsFile)
        return obj.is_file or not obj.is_dir


class IsDirSifter(Sifter):
    def sift(self, obj: Union[PcsFile, str]) -> bool:
        assert isinstance(obj, PcsFile)
        return obj.is_dir or not obj.is_file


T = TypeVar("T", PcsFile, str)


def sift(objs: List[T], sifters: List[Sifter], recursive: bool = False) -> List[T]:
    if sifters:
        obj_dirs: List[T]
        if recursive:
            # If it is recursive, we ignore to sift dirs.
            obj_dirs = [o for o in objs if isinstance(o, PcsFile) and o.is_dir]
            objs = [o for o in objs if not isinstance(o, PcsFile) or isinstance(o, PcsFile) and o.is_file]
        else:
            obj_dirs = []

        objs = obj_dirs + [obj for obj in objs if all([sifter(obj) for sifter in sifters])]
    return objs



# https://stackoverflow.com/a/22085679/2478637
# http://simondlevy.academic.wlu.edu/files/software/kbhit.py

from typing import Callable, Any

import os
import threading
import time

# Windows
if os.name == "nt":
    import msvcrt

# Posix (Linux, OS X)
else:
    import sys
    import termios
    import atexit
    from select import select


class KeyboardListener(threading.Thread):
    def __init__(self, on: Callable[[str], Any]):
        """Creates a KeyboardListener object that you can call to do various keyboard things."""
        super().__init__()

        self._on = on
        self._mt = threading.main_thread()

        if os.name == "nt":
            pass
        else:
            # Save the terminal settings
            self.fd = sys.stdin.fileno()
            self.new_term = termios.tcgetattr(self.fd)
            self.old_term = termios.tcgetattr(self.fd)

            # New terminal setting unbuffered
            self.new_term[3] = self.new_term[3] & ~termios.ICANON & ~termios.ECHO
            termios.tcsetattr(self.fd, termios.TCSAFLUSH, self.new_term)

            # Support normal-terminal reset at exit
            atexit.register(self.set_normal_term)

    def set_normal_term(self):
        """Resets to normal terminal.  On Windows this is a no-op."""

        if os.name == "nt":
            pass
        else:
            termios.tcsetattr(self.fd, termios.TCSAFLUSH, self.old_term)

    def getch(self):
        """Returns a keyboard character after kbhit() has been called.
        Should not be called in the same program as getarrow().
        """

        if os.name == "nt":
            return msvcrt.getch().decode("utf-8")
        else:
            return sys.stdin.read(1)

    def getarrow(self):
        """Returns an arrow-key code after kbhit() has been called. Codes are
        0 : up
        1 : right
        2 : down
        3 : left
        Should not be called in the same program as getch().
        """

        if os.name == "nt":
            msvcrt.getch()  # skip 0xE0
            c = msvcrt.getch()
            vals = [72, 77, 80, 75]
        else:
            c = sys.stdin.read(3)[2]
            vals = [65, 67, 66, 68]

        return vals.index(ord(c.decode("utf-8")))

    def kbhit(self):
        """Returns True if keyboard character was hit, False otherwise."""
        if os.name == "nt":
            return msvcrt.kbhit()

        else:
            dr, dw, de = select([sys.stdin], [], [], 0)
            return dr != []

    def run(self):
        while True:
            if self.kbhit():
                c = self.getch()
                self._on(c)
            else:
                time.sleep(0.1)

            # Exit when main_thread exited
            if not self._mt.is_alive():
                break

        self.set_normal_term()



from os import PathLike
from pathlib import Path

from alipcs_py.common.io import to_decryptio, READ_SIZE
from alipcs_py.common.path import exists


def decrypt_file(from_encrypted: PathLike, to_decrypted: PathLike, encrypt_password: bytes = b""):
    assert exists(from_encrypted)

    dio = to_decryptio(open(from_encrypted, "rb"), encrypt_password=encrypt_password)

    dpath = Path(to_decrypted)
    dir_ = dpath.parent
    if not dir_.exists():
        dir_.mkdir(parents=True)

    with dpath.open("wb") as dfd:
        while True:
            data = dio.read(READ_SIZE)
            if not data:
                break
            dfd.write(data)



import json
import time
import string
import math


def dump_json(obj) -> str:
    return json.dumps(obj, separators=(",", ":"), ensure_ascii=False)


def format_date(timestramp: int) -> str:
    return time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(timestramp))


def format_time(seconds: int) -> str:
    tm = seconds
    ft = ""
    for unit in [60, 60, 24]:
        ft = f"{tm % unit:0>2}" + ft
        if unit != 24:
            ft = ":" + ft
        tm //= unit

    ft = f"{tm}days {ft}"
    return ft


def human_size(size: int) -> str:
    s = float(size)
    v = ""
    t = ""
    for t in ["B", "KB", "MB", "GB", "TB"]:
        if s < 1024.0:
            v = f"{s:3.1f}"
            break
        s /= 1024.0
    if v.endswith(".0"):
        v = v[:-2]
    return f"{v} {t}"


_nums_set = set(string.digits + ".")


def human_size_to_int(size_str: str) -> int:
    size_str = size_str.strip()
    if not size_str:
        return 0

    i = 0
    while i < len(size_str):
        if size_str[i] in _nums_set:
            i += 1
            continue
        else:
            break

    if i == 0:
        return 0

    s = float(size_str[:i])
    _s = s

    unit = size_str[i:].upper().replace(" ", "")
    if not unit:
        return math.floor(_s)

    for t in ["KB", "MB", "GB", "TB"]:
        _s *= 1024
        if unit == t or unit[0] == t[0]:
            return math.floor(_s)

    return math.floor(s)



from typing import cast
import os
from pathlib import Path

from alipcs_py.common.log import LogLevels, TLogLevel, get_logger as _get_logger
from alipcs_py.commands.env import LOG_LEVEL, LOG_PATH


def get_logger(name: str):
    _LOG_PATH = Path(os.getenv("LOG_PATH") or LOG_PATH).expanduser()
    _LOG_LEVEL: TLogLevel = cast(
        TLogLevel,
        os.getenv("LOG_LEVEL", LOG_LEVEL).upper(),
    )

    assert _LOG_LEVEL in LogLevels

    return _get_logger(name, filename=_LOG_PATH, level=_LOG_LEVEL)



from typing import (
    Optional,
    List,
    Tuple,
    Dict,
    Union,
    Any,
    Callable,
    Generator,
    IO,
    cast,
)
from io import BytesIO, UnsupportedOperation
from enum import Enum
from pathlib import Path
from zlib import crc32
from random import Random
import os
import hashlib
import logging
import time

from alipcs_py.common import constant
from alipcs_py.common.number import u64_to_u8x8, u8x8_to_u64
from alipcs_py.common.crypto import (
    random_bytes,
    random_sys_bytes,
    Cryptography,
    SimpleCryptography,
    ChaCha20Cryptography,
    AES256CBCCryptography,
    aes256cbc_encrypt,
    aes256cbc_decrypt,
    padding_key,
    padding_size,
    pkcs7_padding,
    pkcs7_unpadding,
    generate_salt,
    generate_key_iv,
    calc_md5,
    calc_crc32_and_md5,
)
from alipcs_py.common.log import LogLevels, TLogLevel, get_logger

import requests  # type: ignore
from requests import Response

_LOG_LEVEL = os.getenv("LOG_LEVEL", "CRITICAL").upper()
if _LOG_LEVEL not in LogLevels:
    _LOG_LEVEL = "CRITICAL"
logger = get_logger(__name__, level=cast(TLogLevel, _LOG_LEVEL))

READ_SIZE = 65535
DEFAULT_MAX_CHUNK_SIZE = 50 * constant.OneM

BAIDUPCS_PY_CRYPTO_MAGIC_CODE = b"\x00@@#__BAIDUPCS_PY__CRYPTO__#@@\x00\xff"
ENCRYPT_HEAD_LEN = len(BAIDUPCS_PY_CRYPTO_MAGIC_CODE) + 1 + 16 + 8
PADDED_ENCRYPT_HEAD_LEN = padding_size(ENCRYPT_HEAD_LEN, 16)
PADDED_ENCRYPT_HEAD_WITH_SALT_LEN = PADDED_ENCRYPT_HEAD_LEN + 8


def total_len(o: Any) -> int:
    """Read obj len"""

    if hasattr(o, "__len__"):
        return len(o)

    if hasattr(o, "seekable"):
        if o.seekable():
            _offset = o.tell()
            _len = o.seek(0, 2)
            o.seek(_offset, 0)
            return _len

    if hasattr(o, "len"):
        return o.len

    if hasattr(o, "fileno"):
        try:
            fileno = o.fileno()
        except UnsupportedOperation:
            pass
        else:
            return os.fstat(fileno).st_size

    if hasattr(o, "getvalue"):
        # e.g. BytesIO, cStringIO.StringIO
        return len(o.getvalue())

    raise TypeError("Unsupported Operation: no method to get len")


class ChunkIO(IO):
    """Wrap a IO, only read `size` bytes"""

    def __init__(self, io: IO, size: int):
        self._io = io
        self._size = size
        self._offset = 0
        self._io_offset = io.tell()
        self.len = size

    def __len__(self) -> int:
        return self._size

    def read(self, size: int = -1) -> Optional[bytes]:
        remain = self._size - self._offset
        if size == -1:
            size = remain
        else:
            size = min(size, remain)

        data = self._io.read(size) or b""
        if data:
            self._offset += len(data)
        return data

    def seek(self, offset: int, whence: int = 0) -> int:
        if not self.seekable():
            raise ValueError(f"{self._io.__class__.__name__} is not support seeking")

        if whence == 0:
            if offset < 0:
                raise ValueError(f"Negative seek position {offset}")
            pass
        elif whence == 1:
            offset += self._offset
        elif whence == 2:
            offset = self._size - offset

        offset = max(0, offset)
        offset = min(self._size, offset)
        self._io.seek(self._io_offset + offset)
        self._offset = offset
        return self._offset

    def tell(self) -> int:
        return self._offset

    def seekable(self) -> bool:
        return self._io.seekable()

    def writable(self):
        return False

    def close(self):
        pass


def sample_data(io: IO, rg: Random, size: int) -> bytes:
    """Sample data with size"""

    _len = io.seek(0, 2)
    data = b""
    for _ in range(size):
        i = rg.randint(0, max(0, _len - 1))
        io.seek(i, 0)
        data += io.read(1)
    io.seek(0, 0)
    return data


def rapid_upload_params(io: IO) -> Tuple[str, str, int, int]:
    io_len = 0
    chunk_size = constant.OneM
    crc32_v = 0
    md5_v = hashlib.md5()

    buf = io.read(256 * constant.OneK)
    io_len += len(buf)
    slice_md5 = calc_md5(buf)
    md5_v.update(buf)
    t = io.read(chunk_size - 256 * constant.OneK)
    io_len += len(t)
    md5_v.update(t)
    buf += t
    crc32_v = crc32(buf, crc32_v).conjugate()

    # content_crc32, content_md5 = calc_crc32_and_md5(io, constant.OneM)

    while True:
        buf = io.read(chunk_size)
        if buf:
            md5_v.update(buf)
            crc32_v = crc32(buf, crc32_v).conjugate()
            io_len += len(buf)
        else:
            break

    content_crc32, content_md5 = crc32_v.conjugate() & 0xFFFFFFFF, md5_v.hexdigest()

    # if isinstance(io, EncryptIO):
    #     io_len = len(io)
    # else:
    #     io_len = io.seek(0, 2)
    #     io.seek(0, 0)

    return slice_md5, content_md5, content_crc32, io_len


def rapid_upload_params2(localPath: Path) -> Tuple[str, str, int, int]:
    buf = localPath.open("rb").read(constant.OneM)
    slice_md5 = calc_md5(buf[: 256 * constant.OneK])
    content_crc32, content_md5 = calc_crc32_and_md5(localPath.open("rb"), constant.OneM)
    return slice_md5, content_md5, content_crc32, localPath.stat().st_size


def generate_nonce_or_iv(salt: bytes, io: IO) -> bytes:
    io_len = total_len(io)
    rg = Random(salt)
    sample = sample_data(io, rg, 16)
    return random_bytes(16, salt + sample + str(io_len).encode("utf-8"))


class EncryptIO(IO):
    """Encrypt IO

    All methods are implied for stream cipher.
    Block cipher needs to imply by itself.
    """

    def __init__(self, io: IO, encrypt_password: bytes, total_origin_len: int):
        self._io = io

        # The length of unencrypted content of `self._io`
        self._io_len = total_len(self._io)

        # The offset of encrypted content of `self._io`
        self._offset = 0

        # The total length of all original (decrypted) content
        #
        # Only define it at __init__
        self._total_origin_len = total_origin_len

        self._encrypt_password = encrypt_password

        self._salt_for_head = generate_salt()
        self._encrypt_key_for_head, self._iv_for_head = generate_key_iv(encrypt_password, self._salt_for_head, 32, 16)

        # `self._encrypt_key`,`self._nonce_or_iv` is for cryptography
        self._salt = generate_salt()
        self._encrypt_key, self._nonce_or_iv = generate_key_iv(encrypt_password, self._salt, 32, 16)

        # Cryptography
        #
        # Instantiated at each subclass, here is for mypy
        self._crypto: Cryptography

        self._total_head: bytes
        self._total_head_len = len(self.total_head)

    def magic_code(self) -> bytes:
        raise NotImplementedError()

    def block_size(self) -> int:
        raise NotImplementedError()

    def reset(self):
        """Reset io and crypto"""
        self._io.seek(0, 0)
        self._offset = 0
        self._crypto.reset()

    @property
    def total_head(self) -> bytes:
        """
        Version 1:
        aes256cbc_encrypt(`BAIDUPCS_PY_CRYPTO_MAGIC_CODE 32bytes`, encrypt_password, random_bytes(16, encrypt_password)) +
        `encrypt algorithm code 1bytes` +
        `nonce or iv 16bytes` +
        `total_origin_len 8bytes`

        Version 2:
        aes256cbc_encrypt(
            `BAIDUPCS_PY_CRYPTO_MAGIC_CODE 32bytes` +
            `encrypt algorithm code 1bytes` +
            `nonce_or_iv 16bytes`
            `total_origin_len 8bytes`
        ) +
        `salt 8bytes`

        Version 3:
        aes256cbc_encrypt(
            `BAIDUPCS_PY_CRYPTO_MAGIC_CODE 32bytes` +
            `encrypt algorithm code 1bytes` +
            `salt 8bytes` + `random padding 8bytes`
            `total_origin_len 8bytes`
        ) +
        `salt 8bytes`
        """

        if hasattr(self, "_total_head"):
            return self._total_head

        assert len(self._salt) == 8

        ori_head = padding_key(
            BAIDUPCS_PY_CRYPTO_MAGIC_CODE
            + self.magic_code()
            + self._salt
            + random_sys_bytes(8)
            + u64_to_u8x8(self._total_origin_len),
            PADDED_ENCRYPT_HEAD_LEN,
            value=b"",
        )

        # AES256CBC Encrypt head
        self._total_head = (
            aes256cbc_encrypt(ori_head, self._encrypt_key_for_head, self._iv_for_head) + self._salt_for_head
        )
        return self._total_head

    def __len__(self) -> int:
        return self._total_head_len + self._io_len

    def read(self, size: int = -1) -> Optional[bytes]:
        """The read only imply to stream cipher"""

        if self._offset < self._total_head_len:
            if size < 0:
                data = self.total_head[self._offset :]
                io_buf = self._io.read()
                data += self._crypto.encrypt(io_buf)
            else:
                data = self.total_head[self._offset : self._offset + size]
                size -= len(data)
                if size > 0:
                    io_buf = self._io.read(size)
                    data += self._crypto.encrypt(io_buf)
        else:
            io_buf = self._io.read(size)
            data = self._crypto.encrypt(io_buf)
        self._offset += len(data)
        return data

    def seek(self, offset: int, whence: int = 0) -> int:
        if not self.seekable() and offset == 0 and whence == 0 and self._total_head_len > 0:
            self._crypto.reset()
            self._io.seek(0, 0)
            self._offset = 0
            return self._offset

        if not self.seekable():
            raise ValueError(f"{self._crypto.__class__.__name__} is not support seeking")

        if whence == 0:
            if offset < 0:
                raise ValueError(f"Negative seek position {offset}")
            pass
        elif whence == 1:
            offset += self._offset
        elif whence == 2:
            offset = len(self) - offset

        io_offset = offset - self._total_head_len
        if io_offset > 0:
            io_len = self._io.seek(0, 2)
            io_offset = self._io.seek(min(io_offset, io_len), 0)
            self._offset = self._total_head_len + io_offset
        else:
            self._io.seek(0, 0)
            self._offset = max(0, offset)
        return self._offset

    def tell(self) -> int:
        return self._offset

    def seekable(self) -> bool:
        """For stream cipher"""

        return True

    def writable(self) -> bool:
        return False

    def close(self):
        self._io.close()


class SimpleEncryptIO(EncryptIO):
    MAGIC_CODE = b"\x00"

    BLOCK_SIZE = 16  # 16 bytes

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._crypto = SimpleCryptography(self._encrypt_key + self._nonce_or_iv)

    def magic_code(self) -> bytes:
        return self.MAGIC_CODE

    def block_size(self) -> int:
        return self.BLOCK_SIZE


class ChaCha20EncryptIO(EncryptIO):
    MAGIC_CODE = b"\x01"

    BLOCK_SIZE = 16  # 16 bytes

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._crypto = ChaCha20Cryptography(self._encrypt_key, self._nonce_or_iv)

    def magic_code(self) -> bytes:
        return self.MAGIC_CODE

    def block_size(self) -> int:
        return self.BLOCK_SIZE

    def reset(self):
        super().reset()

    def seekable(self) -> bool:
        return False


class AES256CBCEncryptIO(EncryptIO):
    MAGIC_CODE = b"\x02"

    BLOCK_SIZE = 16  # 16 bytes

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._crypto = AES256CBCCryptography(self._encrypt_key, self._nonce_or_iv)

        # The offset of encrypted origin content
        self._origin_io_offset = 0
        # Origin content cache
        self._origin_cache = bytearray()

        # Total length of final encrypted content
        self._encrypted_io_len = padding_size(self._io_len, self.BLOCK_SIZE)
        # Encrypted content cache
        self._encrypted_cache = bytearray()

        self._need_data_padded = self._total_origin_len != self._encrypted_io_len

    def magic_code(self) -> bytes:
        return self.MAGIC_CODE

    def block_size(self) -> int:
        return self.BLOCK_SIZE

    def reset(self):
        super().reset()
        self._origin_io_offset = 0
        self._origin_cache = bytearray()
        self._encrypted_cache = bytearray()

    def __len__(self):
        """Encrypted content length"""

        return self._total_head_len + self._encrypted_io_len

    def _read_block(self, size: int = -1):
        """Read encrypted block to cache"""

        # `self._encrypted_cache` has remains.
        if size > 0 and len(self._encrypted_cache) > size:
            return

        # `size` must be large or equal to the `size` as `size` be the times of `self.BLOCK_SIZE`
        if size > 0:
            size = padding_size(size, self.BLOCK_SIZE)

        data = self._io.read(size) or b""
        if not data:
            return

        self._origin_io_offset += len(data)
        self._origin_cache.extend(data)

        avail_ori_len = padding_size(len(self._origin_cache), self.BLOCK_SIZE, ceil=False)
        if avail_ori_len > 0:
            ori_cn = self._origin_cache[:avail_ori_len]
            self._origin_cache = self._origin_cache[avail_ori_len:]
        else:
            ori_cn = b""

        # The end encryption
        if self._origin_io_offset == self._total_origin_len:
            # Take all remainder
            if self._origin_cache:
                ori_cn += bytes(self._origin_cache)
                self._origin_cache.clear()

            # Padding
            if self._need_data_padded:
                ori_cn = pkcs7_padding(ori_cn, self.BLOCK_SIZE)

            enc_cn = self._crypto.encrypt(ori_cn)
            self._crypto.finalize()
        else:
            enc_cn = self._crypto.encrypt(ori_cn)

        self._encrypted_cache.extend(enc_cn)

    def _takeout(self, size: int = -1) -> bytes:
        """Take out from encrypted cache"""

        self._read_block(size)

        if size < 0:
            data = bytes(self._encrypted_cache)
            self._encrypted_cache.clear()
        else:
            data = bytes(self._encrypted_cache[:size])
            self._encrypted_cache = self._encrypted_cache[size:]
        return data

    def read(self, size: int = -1) -> Optional[bytes]:
        if self._offset < self._total_head_len:
            if size < 0:
                data = self.total_head[self._offset :]
                data += self._takeout()
            else:
                data = self.total_head[self._offset : self._offset + size]
                size -= len(data)
                if size > 0:
                    data += self._takeout(size)
        else:
            data = self._takeout(size)
        self._offset += len(data)
        return data

    def seekable(self) -> bool:
        return False


class DecryptIO(IO):
    def __init__(self, io: IO, encrypt_key: bytes, nonce_or_iv: bytes, total_origin_len: int):
        self._io = io

        # The offset after ENCRYPT_HEAD_LEN
        # Must be equal ENCRYPT_HEAD_LEN
        #
        # When `self.set_io` be called, the `_io_init_offset` must be set to `io.tell()`
        self._io_init_offset = io.tell()

        # The length of decrypted content of `self._io`
        #
        # When `self.set_io` be called, the `_io_len` must be set to new io's len
        self._io_len = total_len(self._io)

        # The offset of decrypted content of `self._io`
        #
        # When `self.set_io` be called, the `_offset` must be set to `io.tell()`
        self._offset = 0

        self._encrypt_key = encrypt_key
        self._nonce_or_iv = nonce_or_iv

        # The total length of all original (decrypted) content
        #
        # Only define it at __init__
        self._total_origin_len = total_origin_len

        self._total_head_len = ENCRYPT_HEAD_LEN

    def __len__(self) -> int:
        return self._io_len - self._io_init_offset

    def set_io(self, io: IO):
        self._io = io
        self._io_init_offset = io.tell()
        self._offset = 0
        self._io_len = total_len(self._io)

    def read(self, size: int = -1) -> Optional[bytes]:
        data = self._io.read(size)
        if not data:
            return b""
        self._offset += len(data)
        return data

    def seek(self, offset: int, whence: int = 0) -> int:
        # ChaCha20 and AES are depended on previous decrypted
        if not self.seekable():
            raise ValueError(f"{self.__class__.__name__} is not support seeking")

        if whence == 0:
            if offset < 0:
                raise ValueError(f"Negative seek position {offset}")
            pass
        elif whence == 1:
            offset += self._offset
        elif whence == 2:
            offset = len(self) - offset

        offset = max(0, offset)
        offset = min(len(self), offset)
        self._io.seek(offset + self._io_init_offset, 0)
        self._offset = offset
        return offset

    def tell(self) -> int:
        return self._offset

    def seekable(self) -> bool:
        return True

    def writable(self) -> bool:
        return False

    def close(self):
        self._io.close()


class SimpleDecryptIO(DecryptIO):
    def __init__(self, io: IO, encrypt_key: bytes, nonce: bytes, total_origin_len: int):
        super().__init__(io, encrypt_key, nonce, total_origin_len)

        self._crypto = SimpleCryptography(self._encrypt_key + self._nonce_or_iv)

    def read(self, size: int = -1) -> Optional[bytes]:
        data = self._io.read(size)
        if not data:
            return b""
        self._offset += len(data)
        return self._crypto.decrypt(data)

    def seekable(self) -> bool:
        return True


class ChaCha20DecryptIO(DecryptIO):
    BLOCK_SIZE = 16  # 16 bytes

    def __init__(self, io: IO, encrypt_key: bytes, nonce: bytes, total_origin_len: int):
        encrypt_key = padding_key(encrypt_key, self.BLOCK_SIZE * 2)
        nonce = padding_key(nonce, self.BLOCK_SIZE)
        super().__init__(io, encrypt_key, nonce, total_origin_len)

        self._crypto = ChaCha20Cryptography(self._encrypt_key, self._nonce_or_iv)

    def read(self, size: int = -1) -> Optional[bytes]:
        data = self._io.read(size)
        if not data:
            return b""
        self._offset += len(data)
        return self._crypto.decrypt(data)

    def seekable(self) -> bool:
        return False


class AES256CBCDecryptIO(DecryptIO):
    BLOCK_SIZE = 16  # 16 bytes

    def __init__(self, io: IO, encrypt_key: bytes, iv: bytes, total_origin_len: int):
        encrypt_key = padding_key(encrypt_key, self.BLOCK_SIZE * 2)
        iv = padding_key(iv, self.BLOCK_SIZE)
        super().__init__(io, encrypt_key, iv, total_origin_len)

        self._crypto = AES256CBCCryptography(self._encrypt_key, self._nonce_or_iv)

        self._encrypted_io_offset = 0
        self._encrypted_io_len = padding_size(total_origin_len, self.BLOCK_SIZE)
        self._encrypted_data_padded = total_origin_len != self._encrypted_io_len

        self._encrypted_cache = bytearray()
        self._decrypted_cache = bytearray()

    def __len__(self) -> int:
        """The decrypted content length of `self._io`"""

        avail_enc_len = padding_size(self._io_len - self._io_init_offset, self.BLOCK_SIZE, ceil=False)
        return avail_enc_len

    def _read_block(self, size: int = -1):
        """Read encrypted block to cache"""

        # `self._decrypted_cache` has remains.
        if size > 0 and len(self._decrypted_cache) > size:
            return

        # `size` must be large or equal to the `size` as `size` be the times of `self.BLOCK_SIZE`
        if size > 0:
            size = padding_size(size, self.BLOCK_SIZE)

        data = self._io.read(size)
        if not data:
            return

        self._encrypted_io_offset += len(data)
        self._encrypted_cache.extend(data)

        avail_enc_len = padding_size(len(self._encrypted_cache), self.BLOCK_SIZE, ceil=False)
        if avail_enc_len > 0:
            enc_cn = bytes(self._encrypted_cache[:avail_enc_len])
            self._encrypted_cache = self._encrypted_cache[avail_enc_len:]
        else:
            enc_cn = b""

        # The end decryption
        if self._encrypted_io_offset == self._encrypted_io_len:
            if self._encrypted_cache:
                enc_cn += bytes(self._encrypted_cache)
                self._encrypted_cache.clear()

            dec_cn = self._crypto.decrypt(enc_cn)

            if self._encrypted_data_padded:
                dec_cn = pkcs7_unpadding(dec_cn, self.BLOCK_SIZE)

            self._crypto.finalize()
        else:
            dec_cn = self._crypto.decrypt(enc_cn)

        self._decrypted_cache.extend(dec_cn)

    def read(self, size: int = -1) -> Optional[bytes]:
        self._read_block(size)

        if size < 0:
            data = bytes(self._decrypted_cache)
            self._decrypted_cache.clear()
        else:
            data = bytes(self._decrypted_cache[:size])
            self._decrypted_cache = self._decrypted_cache[size:]
        self._offset += len(data)
        return data

    def seekable(self) -> bool:
        return False


def parse_head(head: bytes) -> Tuple[bytes, bytes, bytes, bytes]:
    i = len(BAIDUPCS_PY_CRYPTO_MAGIC_CODE)
    return (
        head[:i],
        head[i : i + 1],  # magic_code
        head[i + 1 : i + 1 + 16],  # salt + random padding
        head[i + 1 + 16 : i + 1 + 16 + 8],  # total_origin_len
    )


def _decryptio_version1(total_head: bytes, io: IO, encrypt_password: bytes) -> Optional[DecryptIO]:
    encrypt_password = padding_key(encrypt_password, 32)

    if len(total_head) < ENCRYPT_HEAD_LEN:
        return None

    # Version 1
    b_mc, magic_code, nonce_or_iv, total_origin_len_ = parse_head(total_head)
    b_mc = aes256cbc_decrypt(b_mc, encrypt_password, random_bytes(16, encrypt_password))
    total_origin_len = u8x8_to_u64(total_origin_len_)

    if b_mc != BAIDUPCS_PY_CRYPTO_MAGIC_CODE:
        return None

    if magic_code == SimpleEncryptIO.MAGIC_CODE:
        return SimpleDecryptIO(io, encrypt_password, b"", total_origin_len)
    elif magic_code == ChaCha20EncryptIO.MAGIC_CODE:
        return ChaCha20DecryptIO(io, encrypt_password, nonce_or_iv, total_origin_len)
    elif magic_code == AES256CBCEncryptIO.MAGIC_CODE:
        return AES256CBCDecryptIO(io, encrypt_password, nonce_or_iv, total_origin_len)
    else:
        logging.warning(f"Unknown magic_code: {magic_code!r}")
        return None


def _decryptio_version3(total_head: bytes, io: IO, encrypt_password: bytes) -> Optional[DecryptIO]:
    if len(total_head) < PADDED_ENCRYPT_HEAD_WITH_SALT_LEN:
        return None

    enc_head, salt_for_head = (
        total_head[:PADDED_ENCRYPT_HEAD_LEN],
        total_head[PADDED_ENCRYPT_HEAD_LEN:],
    )

    encrypt_key_for_head, iv_for_head = generate_key_iv(encrypt_password, salt_for_head, 32, 16)

    head = aes256cbc_decrypt(enc_head, encrypt_key_for_head, iv_for_head)

    b_mc, magic_code, padding_salt, total_origin_len_ = parse_head(head)
    total_origin_len = u8x8_to_u64(total_origin_len_)

    salt = padding_salt[:8]
    encrypt_key, nonce_or_iv = generate_key_iv(encrypt_password, salt, 32, 16)

    if b_mc != BAIDUPCS_PY_CRYPTO_MAGIC_CODE:
        return None

    eio: DecryptIO
    if magic_code == SimpleEncryptIO.MAGIC_CODE:
        eio = SimpleDecryptIO(io, encrypt_key, nonce_or_iv, total_origin_len)
    elif magic_code == ChaCha20EncryptIO.MAGIC_CODE:
        eio = ChaCha20DecryptIO(io, encrypt_key, nonce_or_iv, total_origin_len)
    elif magic_code == AES256CBCEncryptIO.MAGIC_CODE:
        eio = AES256CBCDecryptIO(io, encrypt_key, nonce_or_iv, total_origin_len)
    else:
        logging.warning(f"Unknown magic_code: {magic_code!r}")
        return None

    eio._total_head_len = PADDED_ENCRYPT_HEAD_WITH_SALT_LEN
    return eio


def to_decryptio(io: IO, encrypt_password: bytes):
    if not encrypt_password:
        return io

    total_head = io.read(ENCRYPT_HEAD_LEN)
    eio = _decryptio_version1(total_head, io, encrypt_password)
    if eio is not None:
        return eio

    # No support Version 2

    total_head += io.read(PADDED_ENCRYPT_HEAD_WITH_SALT_LEN - ENCRYPT_HEAD_LEN)
    eio = _decryptio_version3(total_head, io, encrypt_password)
    if eio is not None:
        return eio

    io.seek(0, 0)
    return io


class AutoDecryptRequest:
    def __init__(
        self,
        method: str,
        url: str,
        headers: Optional[Dict[str, str]] = None,
        max_chunk_size: int = DEFAULT_MAX_CHUNK_SIZE,
        encrypt_password: bytes = b"",
        **kwargs,
    ):
        kwargs["stream"] = True

        self._method = method
        self._url = url
        self._headers = headers
        self._session = requests.session()
        self._kwargs = kwargs
        self._max_chunk_size = max_chunk_size
        self._encrypt_password = encrypt_password
        self._dio: Optional[IO] = None
        self._has_encrypted = False
        self._total_head_len = 0
        self._decrypted_count = 0
        self._parsed = False

        # Rapid Upload Infos
        #
        # Total raw content length
        self._content_length: Optional[int] = None
        # Total raw content md5
        self._content_hash: Optional[str] = None
        # Total raw content crc32
        self._content_crc64: Optional[int] = None

    def _init(self):
        """Initiate request info"""

        if not self._parsed:
            self._parse_crypto()

        if self._content_length is None:
            # To invoke `self._parse_rapid_upload_info`
            with self._request((0, 1)):
                pass

    @property
    def content_length(self) -> int:
        """Remote content length"""

        self._init()
        assert self._content_length
        return self._content_length

    @property
    def content_hash(self) -> Optional[str]:
        """Remote content length"""

        self._init()
        assert self._content_length
        return self._content_hash

    @property
    def content_crc64(self) -> Optional[int]:
        """Remote content length"""

        self._init()
        assert self._content_length
        return self._content_crc64

    def _parse_rapid_upload_info(self, resp: Response):
        if not resp.ok or self._content_length:
            return

        headers = resp.headers
        if headers.get("Content-Range"):
            self._content_length = int(headers["Content-Range"].split("/")[-1])

        if headers.get("x-oss-hash-value"):
            self._content_hash = headers["x-oss-hash-value"]

        if headers.get("x-oss-hash-crc64ecma"):
            self._content_crc64 = int(headers["x-oss-hash-crc64ecma"])

    def _parse_crypto(self):
        if self._encrypt_password:
            with self._request((0, PADDED_ENCRYPT_HEAD_WITH_SALT_LEN - 1)) as resp:
                raw_data = resp.raw.read()
                if len(raw_data) == PADDED_ENCRYPT_HEAD_WITH_SALT_LEN:
                    self._dio = to_decryptio(BytesIO(raw_data), self._encrypt_password)
                    self._has_encrypted = isinstance(self._dio, DecryptIO)
                    self._total_head_len = cast(DecryptIO, self._dio)._total_head_len if self._has_encrypted else 0
        self._parsed = True

    def _request(self, _range: Tuple[int, int]) -> Response:
        headers = dict(self._headers or {})
        headers["Range"] = "bytes={}-{}".format(*_range)

        while True:
            try:
                resp = self._session.request(self._method, self._url, headers=headers, **self._kwargs)
                if not resp.ok:
                    logger.warning(
                        "`%s._request` request error: status_code: %s, body: %s",
                        self.__class__.__name__,
                        resp.status_code,
                        resp.content[:1000],
                    )

                self._parse_rapid_upload_info(resp)
                break
            except Exception as err:
                logger.warning(
                    "`%s._request` request error: %s",
                    self.__class__.__name__,
                    err,
                )
                raise IOError(f"{self.__class__.__name__} - Request Error") from err

        if not resp.ok:
            raise IOError(
                f"{self.__class__.__name__} - Response is not ok: "
                f"status_code: {resp.status_code}, body: {resp.content[:1000]}"
            )

        return resp

    def rangeable(self) -> bool:
        """Is support uncontinue range request?"""

        self._init()

        if self._dio is None:
            return True
        return self._dio.seekable()

    def __len__(self) -> int:
        """The decrypted data length"""

        self._init()

        assert self._content_length
        return self._content_length - self._total_head_len

    def read(self, _range: Tuple[int, int]) -> Generator[bytes, None, None]:
        self._init()

        start, end = _range
        if end < 0:
            end = len(self)
        end = min(end, len(self))

        if end - start == 0:
            return None

        if start != self._decrypted_count and not self.rangeable():
            raise IndexError(
                "ChaCha20 must decoded data with continue data."
                f"start: {start}, decrypted count: {self._decrypted_count}"
            )

        ranges = self._split_chunk(start + self._total_head_len, end + self._total_head_len)
        for _rg in ranges:
            with self._request(_rg) as resp:
                stream = resp.raw
                while True:
                    buf = stream.read(READ_SIZE)
                    if not buf:
                        break
                    self._decrypted_count += len(buf)
                    bio = BytesIO(buf)
                    if self._dio is not None and isinstance(self._dio, DecryptIO):
                        self._dio.set_io(bio)
                        yield self._dio.read() or b""
                    else:
                        yield buf

    def _split_chunk(self, start: int, end: int) -> List[Tuple[int, int]]:
        """Split the chunks for range header

        Echo chunk has the length at most `self._max_chunk_size`.
        """

        left = list(range(start, end, self._max_chunk_size))
        left.append(end)
        right = [i - 1 for i in left]
        return [(s, e) for s, e in zip(left[:-1], right[1:])]


class RangeRequestIO(IO):
    def __init__(
        self,
        method: str,
        url: str,
        headers: Optional[Dict[str, str]] = None,
        max_chunk_size: int = DEFAULT_MAX_CHUNK_SIZE,
        callback: Callable[..., None] = None,
        encrypt_password: bytes = b"",
        **kwargs,
    ):
        kwargs["stream"] = True

        self._method = method
        self._url = url
        self._headers = headers
        self._kwargs = kwargs
        self._max_chunk_size = max_chunk_size
        self._callback = callback
        self._encrypt_password = encrypt_password

        self.reset()

    def reset(self):
        self._offset = 0
        self._auto_decrypt_request = AutoDecryptRequest(
            self._method,
            self._url,
            headers=self._headers,
            max_chunk_size=self._max_chunk_size,
            encrypt_password=self._encrypt_password,
            **self._kwargs,
        )

    def __len__(self) -> int:
        return len(self._auto_decrypt_request)

    def read(self, size: int = -1) -> Optional[bytes]:
        if size == 0:
            return b""

        if size == -1:
            size = len(self) - self._offset

        start, end = self._offset, self._offset + size

        buf = b""
        for b in self._auto_decrypt_request.read((start, end)):
            buf += b
            self._offset += len(b)
            # Call callback
            if self._callback:
                self._callback(self._offset)
        return buf

    def seek(self, offset: int, whence: int = 0) -> int:
        if whence == 0:
            self._offset = offset
        elif whence == 1:
            self._offset += offset
        elif whence == 2:
            self._offset = len(self) - offset

        self._offset = max(0, self._offset)
        self._offset = min(len(self), self._offset)
        return self._offset

    def tell(self) -> int:
        return self._offset

    def seekable(self) -> bool:
        return self._auto_decrypt_request.rangeable()

    def writable(self) -> bool:
        return False

    def close(self):
        pass


class EncryptType(Enum):
    No = "No"
    Simple = "Simple"
    ChaCha20 = "ChaCha20"
    AES256CBC = "AES256CBC"

    def encrypt_io(self, io: IO, encrypt_password: bytes) -> IO:
        io_len = total_len(io)
        if self == EncryptType.No:
            return io
        elif self == EncryptType.Simple:
            return SimpleEncryptIO(io, encrypt_password, io_len)
        elif self == EncryptType.ChaCha20:
            return ChaCha20EncryptIO(io, encrypt_password, io_len)
        elif self == EncryptType.AES256CBC:
            return AES256CBCEncryptIO(io, encrypt_password, io_len)
        else:
            raise ValueError(f"Unknown EncryptType: {self}")


def reset_encrypt_io(io: Union[IO, EncryptIO]):
    if isinstance(io, EncryptIO):
        io.reset()
    else:
        io.seek(0, 0)



from .pcs import AliPCS, AliOpenPCS, AliOpenAuth
from .api import AliPCSApi, AliPCSApiMix
from .errors import AliPCSError

from .inner import *


__all__ = ["AliPCS", "AliPCSApi", "AliOpenPCS", "AliPCSApiMix", "AliOpenAuth", "AliPCSError"]



from typing import Any
from types import SimpleNamespace
import os
import json

import toml  # type: ignore


def _to_buildin(obj: Any) -> Any:
    if isinstance(obj, SimpleNamespace):
        data = dict(obj.__dict__)
        for field in getattr(obj, "__annotations__", {}):
            val = getattr(obj, field)
            data[field] = _to_buildin(val)
        return data
    elif isinstance(obj, list):
        return [_to_buildin(item) for item in obj]
    elif isinstance(obj, dict):
        return {k: _to_buildin(v) for k, v in obj.items()}
    else:
        return obj


class Share(SimpleNamespace):
    """Share Configuration"""

    store: bool = False


class AppConfig(SimpleNamespace):
    """App Configuration"""

    share: Share = Share()

    @classmethod
    def load(cls, path: str):
        if os.path.exists(path):
            return json.loads(
                json.dumps(toml.load(path)),
                object_hook=lambda d: SimpleNamespace(**d),
            )
        else:
            return cls()

    def dumps(self):
        return toml.dumps(_to_buildin(self))

    def dump(self, path: str):
        with open(path, "w") as f:
            toml.dump(_to_buildin(self), f)



from typing import List, Callable, Any
import sys

from alipcs_py.common.keyboard import KeyboardListener


class KeyHandler:
    def __init__(self, key: str, callback: Callable[..., Any]):
        self._key = key
        self._callback = callback

    def handle(self, key: str):
        if self._key == key:
            self._callback(key)


class KeyboardMonitor:
    KEY_HANDLERS: List[KeyHandler] = []

    @classmethod
    def register(cls, key_handler: KeyHandler):
        cls.KEY_HANDLERS.append(key_handler)

    @classmethod
    def on(cls, key: str):
        for handler in cls.KEY_HANDLERS:
            handler.handle(key)


global _KEYBOARD_LISTENER_STARTED
_KEYBOARD_LISTENER_STARTED = False

global _KEYBOARD_LISTENER
_KEYBOARD_LISTENER = None


def keyboard_listener_start():
    global _KEYBOARD_LISTENER_STARTED
    if _KEYBOARD_LISTENER_STARTED:
        return

    # KeyboardListener is only available in a terminal
    if sys.stdin.isatty():
        listener = KeyboardListener(on=KeyboardMonitor.on)
        listener.start()

        global _KEYBOARD_LISTENER
        _KEYBOARD_LISTENER = listener

    _KEYBOARD_LISTENER_STARTED = True



from typing import Optional, Any
from functools import wraps


class AliPCSError(Exception):
    def __init__(self, message: str, error_code: Optional[str] = None, cause=None):
        self.__cause__ = cause
        self.error_code = error_code
        super().__init__(message)


def parse_error(error_code: str, info: Any = None) -> AliPCSError:
    msg = f"error_code: {error_code}, response: {info}"
    return AliPCSError(msg, error_code=error_code)


def assert_ok(func):
    """Assert the errno of response is not 0"""

    @wraps(func)
    def check(*args, **kwargs):
        info = func(*args, **kwargs)
        error_code = info.get("code")

        if error_code:
            err = parse_error(error_code, str(info))
            raise err

        return info

    return check


def handle_error(func):
    @wraps(func)
    def refresh(*args, **kwargs):
        code = "This is impossible !!!"
        for _ in range(2):
            self = args[0]

            info = func(*args, **kwargs)
            code = info.get("code")
            if code == "AccessTokenInvalid":
                self.refresh()
                continue

            elif code == "ShareLinkTokenInvalid":
                varnames = func.__code__.co_varnames
                idx = varnames.index("share_id")
                if idx < len(args):
                    share_id = args[idx]
                else:
                    share_id = kwargs.get("share_id")

                share_auth = self.__class__.SHARE_AUTHS.get(share_id)
                if share_auth:
                    share_auth.expire_time = 0.0
                continue

            elif code == "DeviceSessionSignatureInvalid":
                self._signature = ""
                continue

            return info

        raise parse_error(code)

    return refresh



from typing import Optional, List, Dict, Any, TYPE_CHECKING
from dataclasses import dataclass
from collections import namedtuple
import time
import re

from alipcs_py.common.date import iso_8601_to_timestamp, now_timestamp

if TYPE_CHECKING:
    from alipcs_py.alipcs.api import AliPCSApi


@dataclass
class PcsRapidUploadInfo:
    """Rapid Upload Info"""

    content_hash: str
    content_length: int
    content_hash_name: str = "sha1"
    crc64_hash: Optional[str] = None
    name: Optional[str] = None

    def _filename(self) -> str:
        return (self.name or "").replace(" ", "%20")

    def ali(self) -> str:
        """ali://<filename>|<content_length>|<content_hash>"""

        filename = self._filename()
        return f"ali://{filename}|{self.content_length}|{self.content_hash}"

    def short(self) -> str:
        """<filename>|<content_length>|<content_hash>"""

        filename = self._filename()
        return f"{filename}|{self.content_length}|{self.content_hash}"

    def all_links(self) -> List[str]:
        return [self.ali(), self.short()]

    @staticmethod
    def hash_link_protocols() -> List[str]:
        return ["ali", "short"]

    @staticmethod
    def default_hash_link_protocol() -> str:
        return "ali"


@dataclass
class PcsFile:
    """
    A Ali PCS file

    path: str  # remote absolute path
    is_dir: Optional[bool] = None
    is_file: Optional[bool] = None
    fs_id: Optional[int] = None  # file id
    size: Optional[int] = None
    md5: Optional[str] = None
    block_list: Optional[List[str]] = None  # block md5 list
    category: Optional[int] = None
    user_id: Optional[int] = None
    created_at: Optional[int] = None  # server created time
    updated_at: Optional[int] = None  # server updated time
    shared: Optional[bool] = None  # this file is shared if True
    """

    file_id: str
    name: str
    parent_file_id: str
    type: str
    is_dir: bool
    is_file: bool
    size: int = 0
    path: str = ""  # remote absolute path

    created_at: Optional[int] = None  # server created time
    updated_at: Optional[int] = None  # server updated time (updated time)

    file_extension: Optional[str] = None
    content_type: Optional[str] = None
    mime_type: Optional[str] = None
    mime_extension: Optional[str] = None
    labels: Optional[List[str]] = None

    status: Optional[str] = None
    hidden: Optional[bool] = None
    starred: Optional[bool] = None
    category: Optional[str] = None
    punish_flag: Optional[int] = None
    encrypt_mode: Optional[str] = None

    drive_id: Optional[str] = None
    domain_id: Optional[str] = None
    upload_id: Optional[str] = None
    async_task_id: Optional[str] = None

    rapid_upload_info: Optional[PcsRapidUploadInfo] = None
    download_url: Optional[str] = None

    @staticmethod
    def from_(info) -> "PcsFile":
        created_at = None
        if "created_at" in info:
            created_at = iso_8601_to_timestamp(info["created_at"])
        updated_at = None
        if "updated_at" in info:
            updated_at = iso_8601_to_timestamp(info["updated_at"])

        rapid_upload_info: Optional[PcsRapidUploadInfo] = None
        if "content_hash" in info:
            rapid_upload_info = PcsRapidUploadInfo(
                content_hash=info.get("content_hash"),
                content_length=info.get("size"),
                content_hash_name=info.get("content_hash_name"),
                crc64_hash=info.get("crc64_hash"),
                name=info.get("name"),
            )

        return PcsFile(
            file_id=info.get("file_id", ""),
            name=info.get("name", ""),
            parent_file_id=info.get("parent_file_id", ""),
            type=info.get("type", ""),
            is_dir=info.get("type") == "folder",
            is_file=info.get("type") == "file",
            size=info.get("size"),
            created_at=created_at,
            updated_at=updated_at,
            file_extension=info.get("file_extension"),
            content_type=info.get("content_type"),
            mime_type=info.get("mime_type"),
            mime_extension=info.get("mime_extension"),
            labels=info.get("labels"),
            status=info.get("status"),
            hidden=info.get("hidden"),
            starred=info.get("starred"),
            category=info.get("category"),
            punish_flag=info.get("punish_flag"),
            encrypt_mode=info.get("encrypt_mode"),
            drive_id=info.get("drive_id"),
            domain_id=info.get("domain_id"),
            upload_id=info.get("upload_id"),
            async_task_id=info.get("async_task_id"),
            rapid_upload_info=rapid_upload_info,
            download_url=info.get("download_url"),
        )

    @staticmethod
    def root() -> "PcsFile":
        return PcsFile(
            file_id="root",
            name="",
            parent_file_id="root",
            type="folder",
            is_dir=True,
            is_file=False,
            size=0,
            path="/",
        )

    def is_root(self) -> bool:
        return self.file_id == "root"

    def download_url_expires(self) -> bool:
        """Check whether the `self.download_url` expires"""

        if self.download_url:
            mod = re.search(r"oss-expires=(\d+)", self.download_url)
            if mod:
                expire_time = float(mod.group(1))
                if time.time() < expire_time - 5:
                    return False
        return True

    def update_download_url(self, api: "AliPCSApi"):
        """Update the download url if it expires"""

        if self.is_file:
            if self.download_url_expires():
                pcs_url = api.download_link(self.file_id)
                if pcs_url:
                    self.download_url = pcs_url.url
            else:
                if getattr(api, "_aliopenpcsapi", None):
                    pcs_url = api.download_link(self.file_id)
                    if pcs_url:
                        self.download_url = pcs_url.url


@dataclass
class PcsUploadUrl:
    upload_url: Optional[str] = None
    internal_upload_url: Optional[str] = None
    content_type: Optional[str] = None
    part_number: Optional[int] = None

    @staticmethod
    def from_(info) -> "PcsUploadUrl":
        return PcsUploadUrl(
            upload_url=info.get("upload_url"),
            internal_upload_url=info.get("internal_upload_url"),
            content_type=info.get("content_type"),
            part_number=info.get("part_number"),
        )


@dataclass
class PcsPreparedFile:
    file_id: Optional[str] = None
    parent_file_id: Optional[str] = None
    file_name: Optional[str] = None
    type: Optional[str] = None
    part_info_list: Optional[List[PcsUploadUrl]] = None

    rapid_upload: Optional[bool] = None
    pre_hash: Optional[str] = None

    domain_id: Optional[str] = None
    drive_id: Optional[str] = None
    upload_id: Optional[str] = None

    encrypt_mode: Optional[str] = None
    location: Optional[str] = None

    @staticmethod
    def from_(info) -> "PcsPreparedFile":
        part_info_list = [PcsUploadUrl.from_(i) for i in info.get("part_info_list") or []]
        return PcsPreparedFile(
            file_id=info.get("file_id"),
            parent_file_id=info.get("parent_file_id"),
            file_name=info.get("file_name"),
            type=info.get("type"),
            part_info_list=part_info_list,
            rapid_upload=info.get("rapid_upload"),
            pre_hash=info.get("pre_hash"),
            domain_id=info.get("domain_id"),
            drive_id=info.get("drive_id"),
            upload_id=info.get("upload_id"),
            encrypt_mode=info.get("encrypt_mode"),
            location=info.get("location"),
        )

    def can_rapid_upload(self) -> bool:
        return bool(self.pre_hash)

    def is_rapid_upload(self) -> bool:
        return bool(self.rapid_upload)

    def upload_urls(self, internal: bool = False) -> List[str]:
        if internal:
            return [p.internal_upload_url for p in self.part_info_list or [] if p.internal_upload_url]
        else:
            return [p.upload_url for p in self.part_info_list or [] if p.upload_url]


@dataclass
class PcsSharedLink:
    """The shared link by the signin user"""

    share_url: str
    file_id: str
    file_id_list: List[str]
    share_id: str
    share_name: Optional[str] = None
    share_policy: Optional[str] = None
    share_pwd: Optional[str] = None
    drive_id: Optional[str] = None

    # The expiration timestramp
    expiration: Optional[int] = None
    expired: Optional[bool] = None

    # created_at
    created_at: Optional[int] = None
    # updated_at
    updated_at: Optional[int] = None

    status: Optional[str] = None
    preview_count: Optional[int] = None
    save_count: Optional[int] = None
    download_count: Optional[int] = None

    description: Optional[str] = None
    creator: Optional[str] = None

    @staticmethod
    def from_(info) -> "PcsSharedLink":
        created_at = None
        if "created_at" in info:
            created_at = iso_8601_to_timestamp(info["created_at"])
        updated_at = None
        if "updated_at" in info:
            updated_at = iso_8601_to_timestamp(info["updated_at"])
        expiration = None
        if info.get("expiration"):
            expiration = iso_8601_to_timestamp(info["expiration"])

        return PcsSharedLink(
            share_url=info["share_url"],
            file_id=info["file_id"],
            file_id_list=info["file_id_list"],
            share_id=info["share_id"],
            share_name=info.get("share_name"),
            share_policy=info.get("share_policy"),
            share_pwd=info.get("share_pwd"),
            drive_id=info.get("drive_id"),
            expiration=expiration,
            expired=info.get("expired"),
            created_at=created_at,
            updated_at=updated_at,
            status=info.get("status"),
            preview_count=info.get("preview_count"),
            save_count=info.get("save_count"),
            download_count=info.get("download_count"),
            description=info.get("description"),
            creator=info.get("creator"),
        )

    def has_password(self) -> bool:
        return bool(self.share_pwd)


@dataclass
class PcsSharedLinkInfo:
    """The shared link by anyone"""

    share_id: str
    share_pwd: Optional[str] = None
    share_name: Optional[str] = None
    display_name: Optional[str] = None
    file_count: Optional[str] = None
    file_infos: Optional[str] = None
    expiration: Optional[int] = None
    updated_at: Optional[int] = None
    vip: Optional[str] = None
    avatar: Optional[str] = None
    is_following_creator: Optional[str] = None
    creator_id: Optional[str] = None
    creator_name: Optional[str] = None
    creator_phone: Optional[str] = None

    @staticmethod
    def from_(info) -> "PcsSharedLinkInfo":
        updated_at = None
        if "updated_at" in info:
            updated_at = iso_8601_to_timestamp(info["updated_at"])
        expiration = None
        if info.get("expiration"):
            expiration = iso_8601_to_timestamp(info["expiration"])

        return PcsSharedLinkInfo(
            share_id=info.get("share_id"),
            share_pwd=info.get("share_pwd"),
            share_name=info.get("share_name"),
            display_name=info.get("display_name"),
            file_count=info.get("file_count"),
            file_infos=info.get("file_infos"),
            expiration=expiration,
            updated_at=updated_at,
            vip=info.get("vip"),
            avatar=info.get("avatar"),
            is_following_creator=info.get("is_following_creator"),
            creator_id=info.get("creator_id"),
            creator_name=info.get("creator_name"),
            creator_phone=info.get("creator_phone"),
        )

    @property
    def share_url(self) -> str:
        return "https://www.aliyundrive.com/s/" + self.share_id


@dataclass
class SharedAuth:
    """The shared link auth by anyone"""

    share_id: str
    share_password: str
    share_token: str
    expire_time: float
    expires_in: int
    info: Any

    @staticmethod
    def from_(share_id: str, share_password: str, info: Any) -> "SharedAuth":
        return SharedAuth(
            share_id,
            share_password,
            share_token=info["share_token"],
            expire_time=iso_8601_to_timestamp(info["expire_time"]),
            expires_in=info["expires_in"],
            info=info,
        )

    def is_expired(self) -> bool:
        return time.time() >= self.expire_time


FromTo = namedtuple("FromTo", ["from_", "to_"])


@dataclass
class PcsSpace:
    used_size: int
    total_size: int

    @staticmethod
    def from_(info) -> "PcsSpace":
        return PcsSpace(
            used_size=info.get("used_size"),
            total_size=info.get("total_size"),
        )


@dataclass
class PcsAuth:
    refresh_token: str
    access_token: Optional[str] = None


@dataclass
class PcsUserPrivilege:
    feature_id: Optional[str] = None
    feature_attr_id: Optional[str] = None
    quota: Optional[int] = None

    @staticmethod
    def from_(info) -> "PcsUserPrivilege":
        return PcsUserPrivilege(
            feature_id=info.get("feature_id"),
            feature_attr_id=info.get("feature_attr_id"),
            quota=info.get("quota"),
        )


@dataclass
class PcsUserRights:
    spu_id: Optional[str] = None
    name: Optional[str] = None
    is_expires: Optional[str] = None
    privileges: Optional[List[PcsUserPrivilege]] = None

    @staticmethod
    def from_(info) -> "PcsUserRights":
        privileges = [PcsUserPrivilege.from_(i) for i in info.get("privileges") or []]
        return PcsUserRights(
            spu_id=info.get("spu_id"),
            name=info.get("name"),
            is_expires=info.get("is_expires"),
            privileges=privileges,
        )


@dataclass
class PcsUserVip:
    identity: str
    icon: str
    status: str
    vip_list: List[Dict]

    @staticmethod
    def from_(info) -> "PcsUserVip":
        return PcsUserVip(
            identity=info.get("identity"),
            icon=info.get("icon"),
            status=info.get("status"),
            vip_list=info.get("vipList"),
        )

    def is_vip(self) -> bool:
        return self.identity == "vip"


@dataclass
class PcsUser:
    user_id: str
    default_drive_id: Optional[str] = None
    device_id: Optional[str] = None
    domain_id: Optional[str] = None

    web_refresh_token: Optional[str] = None
    web_access_token: Optional[str] = None
    web_token_type: Optional[str] = None
    web_expire_time: Optional[int] = None

    openapi_refresh_token: Optional[str] = None
    openapi_access_token: Optional[str] = None
    openapi_token_type: Optional[str] = None
    openapi_expire_time: Optional[int] = None

    client_id: Optional[str] = None
    client_secret: Optional[str] = None
    client_server: Optional[str] = None

    user_name: Optional[str] = None
    nick_name: Optional[str] = None
    user_data: Optional[Dict[Any, Any]] = None
    phone: Optional[str] = None
    email: Optional[str] = None
    avatar: Optional[str] = None
    personal_space_info: Optional[PcsSpace] = None
    personal_rights_info: Optional[PcsUserRights] = None
    user_vip_info: Optional[PcsUserVip] = None

    created_at: Optional[int] = None
    updated_at: Optional[int] = None

    role: Optional[str] = None
    status: Optional[str] = None
    description: Optional[str] = None

    deny_change_password_by_self: Optional[str] = None
    need_change_password_next_login: Optional[str] = None

    @staticmethod
    def from_(info) -> "PcsUser":
        created_at = None
        if "created_at" in info:
            created_at = info["created_at"] // 1000
        updated_at = None
        if "updated_at" in info:
            updated_at = info["updated_at"] // 1000

        return PcsUser(
            user_id=info["user_id"],
            default_drive_id=info.get("default_drive_id"),
            device_id=info.get("device_id"),
            domain_id=info.get("domain_id"),
            user_name=info.get("user_name"),
            nick_name=info.get("nick_name"),
            user_data=info.get("user_data"),
            phone=info.get("phone"),
            email=info.get("email"),
            avatar=info.get("avatar"),
            personal_space_info=PcsSpace.from_(info.get("personal_space_info") or {}),
            personal_rights_info=PcsUserRights.from_(info.get("personal_rights_info") or {}),
            user_vip_info=PcsUserVip.from_(info.get("user_vip_info") or {}),
            created_at=created_at,
            updated_at=updated_at,
            role=info.get("role"),
            status=info.get("status"),
            description=info.get("description"),
            deny_change_password_by_self=info.get("deny_change_password_by_self"),
            need_change_password_next_login=info.get("need_change_password_next_login"),
        )


@dataclass
class PcsRateLimit:
    part_speed: Optional[int] = None
    part_size: Optional[int] = None

    @staticmethod
    def from_(info) -> "PcsRateLimit":
        return PcsRateLimit(
            part_speed=info.get("part_speed"),
            part_size=info.get("part_size"),
        )


@dataclass
class PcsDownloadUrl:
    url: Optional[str] = None
    internal_url: Optional[str] = None
    cdn_url: Optional[str] = None
    size: Optional[int] = None

    method: Optional[str] = None
    expiration: Optional[int] = None
    ratelimit: Optional[PcsRateLimit] = None

    @staticmethod
    def from_(info) -> "PcsDownloadUrl":
        expiration = None
        if info.get("expiration"):
            expiration = iso_8601_to_timestamp(info["expiration"])

        return PcsDownloadUrl(
            url=info.get("url"),
            internal_url=info.get("internal_url"),
            cdn_url=info.get("cdn_url"),
            size=info.get("size"),
            method=info.get("method"),
            expiration=expiration,
            ratelimit=PcsRateLimit.from_(info.get("ratelimit") or {}),
        )


@dataclass
class AuthInfo:
    refresh_token: str
    access_token: str
    token_type: str
    expire_time: int

    @staticmethod
    def from_(info) -> "AuthInfo":
        expire_time = now_timestamp() + info.get("expires_in", 7200)
        return AuthInfo(
            refresh_token=info.get("refresh_token"),
            access_token=info.get("access_token"),
            token_type=info.get("token_type"),
            expire_time=expire_time,
        )



from typing import Optional, Dict, List, Union, Any, Callable, IO
from enum import Enum
import threading
from urllib.parse import urljoin
from typing_extensions import Literal

import requests  # type: ignore
from requests_toolbelt import MultipartEncoder, MultipartEncoderMonitor

from alipcs_py.common.date import now_timestamp, iso_8601_to_timestamp, timestamp_to_iso_8601
from alipcs_py.common import constant
from alipcs_py.common.io import RangeRequestIO, DEFAULT_MAX_CHUNK_SIZE, ChunkIO, total_len
from alipcs_py.common.cache import timeout_cache
from alipcs_py.common.crypto import generate_secp256k1_keys
from alipcs_py.alipcs.errors import AliPCSError, parse_error, handle_error
from alipcs_py.alipcs.errors import assert_ok
from alipcs_py.alipcs.inner import SharedAuth


UPLOAD_CHUNK_SIZE = 10 * constant.OneM

APP_ID = "5dde4e1bdf9e4966b387ba58f4b3fdc3"

ALIYUNDRIVE_COM = "https://www.aliyundrive.com"
ALIYUNDRIVE_COM_API = "https://api.aliyundrive.com"
ALIYUNDRIVE_OPENAPI_DOMAIN = "https://openapi.aliyundrive.com"

PCS_UA = "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/110.0.0.0 Safari/537.36"
PCS_HEADERS = {"Origin": ALIYUNDRIVE_COM[:-1], "Referer": ALIYUNDRIVE_COM + "/", "User-Agent": PCS_UA}

CheckNameMode = Literal[
    "overwrite",  # 直接覆盖，以后多版本有用
    "auto_rename",  # 自动换一个随机名称
    "refuse",  # 不会创建，告诉你已经存在
    "ignore",  # 会创建重名的
]


class Method(Enum):
    Head = "HEAD"
    Get = "GET"
    Post = "POST"


class PcsNode(Enum):
    f"""Aliyundrive PCS Nodes which use {ALIYUNDRIVE_COM_API}"""

    Token = "v2/account/token"
    Refresh = "token/refresh"
    CreateSession = "users/v1/users/device/create_session"

    FileList = "adrive/v3/file/list"
    Meta = "v2/file/get"
    Search = "adrive/v3/file/search"
    DownloadUrl = "v2/file/get_download_url"

    CreateFile = "v2/file/create"
    UploadComplete = "v2/file/complete"

    Batch = "v3/batch"

    CreateWithFolders = "adrive/v2/file/createWithFolders"
    Move = "adrive/v3/file/move"
    Update = "v3/file/update"
    Trash = "v2/recyclebin/trash"

    Available = "adrive/v2/share_link/check_avaliable"
    Share = "adrive/v2/share_link/create"
    ShareToken = "v2/share_link/get_share_token"
    SharedInfo = "adrive/v3/share_link/get_share_by_anonymous"
    SharedList = "adrive/v3/share_link/list"
    SharedFileDownloadUrl = "v2/file/get_share_link_download_url"

    PersonalInfo = "v2/databox/get_personal_info"
    User = "v2/user/get"
    UserVip = "business/v1.0/users/vip/info"

    def url(self) -> str:
        return f"{ALIYUNDRIVE_COM_API}/{self.value}"


# The Lock is used to refresh token
_LOCK = threading.Lock()


class AliPCS:
    """`AliPCS` provides pcs's apis which return raw json"""

    SHARE_AUTHS: Dict[str, SharedAuth] = {}

    def __init__(
        self,
        refresh_token: str,
        access_token: str = "",
        token_type: str = "Bearer",
        expire_time: int = 0,
        user_id: str = "",
        user_name: str = "",
        nick_name: str = "",
        device_id: str = "",
        default_drive_id: str = "",
        role: str = "",
        status: str = "",
    ):
        self._session = requests.Session()
        self._refresh_token = refresh_token
        self._access_token = access_token
        self._token_type = token_type
        self._expire_time = expire_time

        self._user_id = user_id
        self._user_name = user_name
        self._nick_name = nick_name
        self._device_id = device_id
        self._signature = ""
        self._default_drive_id = default_drive_id
        self._role = role
        self._status = status
        self._nonce = 0

    def __str__(self) -> str:
        return f"""AliPCS

    user_id: {self.user_id}
    user_name: {self.user_name}
    nick_name: {self.nick_name}

    refresh_token: {self.refresh_token}
    access_token: {self.access_token}
    token_type: {self.token_type}
    expire_time: {self.expire_time}

    device_id: {self.device_id}
    signature: {self.signature}
    default_drive_id: {self.default_drive_id}
    role: {self.role}
    status: {self.status}
    nonce: {self._nonce}
    """

    def get_token(self):
        """Get authorization token"""

        url = PcsNode.Token.url()
        data = dict(refresh_token=self.refresh_token, grant_type="refresh_token")
        resp = self._request(Method.Post, url, json=data)
        return resp.json()

    @property
    def refresh_token(self) -> str:
        with _LOCK:
            if not self._access_token or (now_timestamp() - 1 * 60 * 60) >= self._expire_time or not self._device_id:
                self.refresh()
            return self._refresh_token

    @property
    def access_token(self) -> str:
        self.refresh_token
        return self._access_token

    @property
    def user_id(self) -> str:
        self.refresh_token
        return self._user_id

    @property
    def user_name(self) -> str:
        self.refresh_token
        return self._user_name

    @property
    def nick_name(self) -> str:
        self.refresh_token
        return self._nick_name

    @property
    def token_type(self) -> str:
        self.refresh_token
        return self._token_type

    @property
    def expire_time(self) -> int:
        self.refresh_token
        return self._expire_time

    @property
    def device_id(self) -> str:
        self.refresh_token
        return self._device_id

    @property
    def signature(self) -> str:
        if not self._signature:
            self.create_session()
        return self._signature

    @property
    def default_drive_id(self) -> str:
        self.refresh_token
        return self._default_drive_id

    @property
    def role(self) -> str:
        self.refresh_token
        return self._role

    @property
    def status(self) -> str:
        self.refresh_token
        return self._status

    def _request(
        self,
        method: Method,
        url: str,
        params: Optional[Dict[str, str]] = None,
        headers: Optional[Dict[str, str]] = None,
        data: Union[str, bytes, Dict[str, str], Any] = None,
        json: Any = None,
        files: Optional[Dict[str, Any]] = None,
        refresh: bool = False,
        **kwargs,
    ) -> requests.Response:
        if not headers:
            headers = dict(PCS_HEADERS)

        if not refresh and "Authorization" not in headers:
            headers["Authorization"] = f"{self.token_type} {self.access_token}"
            headers["x-device-id"] = self.device_id

        if json is not None:
            headers["Content-Type"] = "application/json;charset=UTF-8"

        if isinstance(data, (MultipartEncoder, MultipartEncoderMonitor)):
            headers["Content-Type"] = data.content_type

        headers["x-canary"] = "client=web,app=adrive,version=v4.1.0"

        try:
            resp = self._session.request(
                method.value, url, params=params, headers=headers, data=data, json=json, files=files, **kwargs
            )
            return resp
        except Exception as err:
            raise AliPCSError("AliPCS._request", cause=err)

    def refresh(self):
        """Refresh token"""

        url = PcsNode.Refresh.url()
        data = dict(refresh_token=self._refresh_token)
        resp = self._request(Method.Post, url, json=data, refresh=True)
        info = resp.json()

        if "code" in info:
            err = parse_error(info["code"], info=info)
            raise err

        self._user_id = info["user_id"]
        self._user_name = info["user_name"]
        self._nick_name = info["nick_name"]

        self._refresh_token = info["refresh_token"]
        self._access_token = info["access_token"]
        self._token_type = info["token_type"]
        self._expire_time = iso_8601_to_timestamp(info["expire_time"])

        self._device_id = info["device_id"]
        self._default_drive_id = info["default_drive_id"]

        self._role = info["role"]
        self._status = info["status"]

        return info

    def create_session(self):
        """Create session"""

        assert self.device_id, "No AliPCS._device_id"
        assert self.user_id, "No AliPCS._user_id"

        device_id = self.device_id
        user_id = self.user_id
        nonce = self._nonce

        private_key, public_key = generate_secp256k1_keys()
        message = f"{APP_ID}:{device_id}:{user_id}:{nonce}".encode()
        signature: str = private_key.sign(message).hex()
        sign = signature + "00"

        url = PcsNode.CreateSession.url()
        data = dict(deviceName="Chrome浏览器", modelName="Mac OS网页版", pubKey=public_key.to_string().hex())

        headers = dict(**PCS_HEADERS)
        headers.update({"x-device-id": device_id, "x-signature": sign})

        @handle_error
        def do(self):
            resp = self._request(Method.Post, url, headers=headers, json=data)
            return resp.json()

        info = do(self)
        if info["result"] and info["success"]:
            self._signature = sign

        return info

    def meta(self, *file_ids: str, share_id: str = None):
        assert "root" not in file_ids, '"root" has NOT meta info'

        headers = dict(PCS_HEADERS)
        headers.update({"x-device-id": self.device_id, "x-signature": self.signature})

        if share_id:
            share_token = self.share_token(share_id)
            assert share_token, "Need share_token"

            headers["x-share-token"] = share_token

        responses = []
        for file_id in file_ids:
            data = dict(
                file_id=file_id,
                fields="*",
                image_thumbnail_process="image/resize,w_400/format,jpeg",
                image_url_process="image/resize,w_375/format,jpeg",
                video_thumbnail_process="video/snapshot,t_1000,f_jpg,ar_auto,w_375",
            )

            if share_id:
                data["share_id"] = share_id
            else:
                data["drive_id"] = self.default_drive_id

            url = PcsNode.Meta.url()
            resp = self._request(Method.Post, url, headers=headers, json=data)
            info = resp.json()
            responses.append(dict(body=info))

        return dict(responses=responses)

    def exists(self, file_id: str) -> bool:
        if file_id == "root":
            return True

        r = self.meta(file_id)
        info = r["responses"][0]["body"]
        if info.get("code") == "NotFound.File":
            return False
        else:
            return True

    def is_file(self, file_id: str) -> bool:
        if file_id == "root":
            return False

        r = self.meta(file_id)
        info = r["responses"][0]["body"]
        if info.get("code") == "NotFound.File":
            return False
        if info["type"] == "file":
            return True
        else:
            return False

    def is_dir(self, file_id: str) -> bool:
        if file_id == "root":
            return True

        r = self.meta(file_id)
        info = r["responses"][0]["body"]
        if info.get("code") == "NotFound.File":
            return False
        if info["type"] == "folder":
            return True
        else:
            return False

    @assert_ok
    @handle_error
    def list(
        self,
        file_id: str,
        share_id: str = "",
        desc: bool = False,
        name: bool = False,
        time: bool = False,
        size: bool = False,
        all: bool = False,
        limit: int = 200,
        url_expire_sec: int = 14400,
        next_marker: str = "",
    ):
        """List files at the directory which has `file_id`

        Args:
            file_id (str):
                The file_id of the directory
        """

        assert limit <= 200, "`limit` should be less than 200"

        url = PcsNode.FileList.url()
        orderby = "name"
        if name:
            orderby = "name"
        elif time:
            orderby = "updated_at"  # "created_at"  # 服务器最后修改时间
        elif size:
            orderby = "size"

        data = dict(
            all=all,
            drive_id=self.default_drive_id,
            fields="*",
            limit=limit,
            order_by=orderby,
            order_direction="DESC" if desc else "ASC",
            parent_file_id=file_id,
            url_expire_sec=url_expire_sec,
            image_thumbnail_process="image/resize,w_256/format,jpeg",
            image_url_process="image/resize,w_1920/format,jpeg/interlace,1",
            video_thumbnail_process="video/snapshot,t_1000,f_jpg,ar_auto,w_256",
        )

        if next_marker:
            data["marker"] = next_marker

        headers = dict(PCS_HEADERS)
        headers.update({"x-device-id": self.device_id, "x-signature": self.signature})
        if share_id:
            share_token = self.share_token(share_id)
            assert share_token, "Need share_token"

            data["share_id"] = share_id
            data.pop("drive_id")
            headers["x-share-token"] = share_token

        resp = self._request(Method.Post, url, headers=headers, json=data)
        return resp.json()

    @staticmethod
    def part_info_list(part_number: int) -> List[Dict[str, int]]:
        return [dict(part_number=i) for i in range(1, part_number + 1)]

    @handle_error
    def create_file(
        self,
        filename: str,
        dir_id: str,
        size: int,
        pre_hash: str = "",
        content_hash: str = "",
        part_number: int = 1,
        proof_code: str = "",
        check_name_mode: CheckNameMode = "auto_rename",
    ):
        """Create a prepared file for uploading

        filename (str):
            The name of file.
        dir_id (str):
            The directory id where the file is at.
        size (int):
            the length of total content.
        pre_hash (str):
            The sha1 of the IO first 1k bytes
        content_hash (str):
            the sha1 of total content.
        part_number (int):
            The number of one file's chunks to upload.
            The server will returns the number of urls to prepare to upload the file's chunks.
            `WARNNING`: this value MUST be set by caller.
        check_name_mode(str):
            'overwrite' (直接覆盖，以后多版本有用)
            'auto_rename' (自动换一个随机名称)
            'refuse' (不会创建，告诉你已经存在)
            'ignore' (会创建重名的)
        """

        url = PcsNode.CreateWithFolders.url()
        data = dict(
            drive_id=self.default_drive_id,
            part_info_list=self.part_info_list(part_number),
            parent_file_id=dir_id,
            name=filename,
            type="file",
            check_name_mode=check_name_mode,
            size=size,
            pre_hash=pre_hash,
            content_hash=content_hash,
            content_hash_name="sha1",
            proof_code=proof_code,
            proof_version="v1",
        )

        resp = self._request(Method.Post, url, json=data)
        return resp.json()

    def prepare_file(self, filename: str, dir_id: str, size: int, pre_hash: str = "", part_number: int = 1):
        return self.create_file(filename, dir_id, size, pre_hash=pre_hash, part_number=part_number)

    @assert_ok
    @handle_error
    def rapid_upload_file(self, filename: str, dir_id: str, size: int, content_hash: str, proof_code: str):
        """Rapid Upload File

        size (int): the length of total content.
        content_hash (str): the sha1 of total content.
        """

        return self.create_file(filename, dir_id, size, content_hash=content_hash, proof_code=proof_code)

    def upload_slice(self, io: IO, url: str, callback: Callable[[MultipartEncoderMonitor], None] = None) -> None:
        """Upload the content of io to remote url"""

        cio = ChunkIO(io, total_len(io))
        monitor = MultipartEncoderMonitor(cio, callback=callback)

        session = requests.Session()
        session.request(
            "PUT",
            url,
            headers=dict(PCS_HEADERS),
            data=monitor,
            # timeout=(3, 9),  # (connect timeout, read timeout)
        )

    @assert_ok
    @handle_error
    def upload_complete(self, file_id: str, upload_id: str):
        url = PcsNode.UploadComplete.url()
        data = dict(drive_id=self.default_drive_id, file_id=file_id, upload_id=upload_id)
        resp = self._request(Method.Post, url, json=data)
        return resp.json()

    @assert_ok
    @handle_error
    def search(
        self,
        keyword: str,
        desc: bool = False,
        name: bool = False,
        time: bool = False,
        size: bool = False,
        all: bool = False,
        limit: int = 100,
        next_marker: str = "",
    ):
        assert limit <= 100, "`limit` should be less than 100"

        url = PcsNode.Search.url()
        orderby = "name"
        if name:
            orderby = "name"
        elif time:
            orderby = "updated_at"  # "created_at"  # 服务器最后修改时间
        elif size:
            orderby = "size"

        data = dict(
            all=all,
            drive_id=self.default_drive_id,
            limit=limit,
            order_by=orderby + " " + ("DESC" if desc else "ASC"),
            query=f'name match "{keyword}"',
            marker=next_marker,
            image_thumbnail_process="image/resize,w_160/format,jpeg",
            image_url_process="image/resize,w_1920/format,jpeg",
            video_thumbnail_process="video/snapshot,t_0,f_jpg,ar_auto,w_300",
        )
        resp = self._request(Method.Post, url, json=data)
        return resp.json()

    @assert_ok
    @handle_error
    def makedir(self, dir_id: str, name: str):
        url = PcsNode.CreateWithFolders.url()
        data = dict(
            check_name_mode="refuse",  # or "auto_rename"
            drive_id=self.default_drive_id,
            name=name,
            parent_file_id=dir_id,
            type="folder",
        )
        resp = self._request(Method.Post, url, json=data)
        return resp.json()

    @assert_ok
    @handle_error
    def batch_operate(
        self, requests_: List[Dict[str, Any]], resource: str = "file", headers: Optional[Dict[str, str]] = None
    ):
        url = PcsNode.Batch.url()
        data = dict(resource=resource, requests=requests_)
        resp = self._request(Method.Post, url, headers=headers, json=data)
        return resp.json()

    def move(self, *file_ids: str):
        """
        Move sources to destination

        source_ids, dest_id = file_ids[:-1], file_ids[-1]

        `dest_id` must be a directory
        """

        source_ids, dest_id = file_ids[:-1], file_ids[-1]

        if self.is_file(dest_id):
            raise AliPCSError("The remote `dest_id` is a file. It must be a directory.")

        requests_ = []
        for source_id in source_ids:
            req = dict(
                method="POST",
                url="/file/move",
                id=source_id,
                headers={"Content-Type": "application/json"},
                body=dict(
                    drive_id=self.default_drive_id,
                    to_drive_id=self.default_drive_id,
                    file_id=source_id,
                    to_parent_file_id=dest_id,
                ),
            )
            requests_.append(req)

        return self.batch_operate(requests_, resource="file")

    @assert_ok
    @handle_error
    def rename(self, file_id: str, name: str):
        """Rename the file to `name`"""

        url = PcsNode.Update.url()
        data = dict(
            check_name_mode="refuse", drive_id=self.default_drive_id, file_id=file_id, name=name  # or "auto_rename"
        )
        resp = self._request(Method.Post, url, json=data)
        return resp.json()

    def copy(self, *file_ids: str):
        assert len(file_ids) > 1

        to_dir_id = file_ids[-1]
        requests_ = []
        for file_id in file_ids[:-1]:
            req = dict(
                method="POST",
                url="/file/copy",
                id=file_id,
                headers={"Content-Type": "application/json"},
                body=dict(
                    drive_id=self.default_drive_id,
                    file_id=file_id,
                    to_parent_file_id=to_dir_id,
                    overwrite=False,
                    auto_rename=True,
                ),
            )
            requests_.append(req)
        return self.batch_operate(requests_, resource="file")

    def remove(self, *file_ids: str):
        requests_ = []
        for file_id in file_ids:
            req = dict(
                method="POST",
                url="/recyclebin/trash",
                id=file_id,
                headers={"Content-Type": "application/json"},
                body=dict(drive_id=self.default_drive_id, file_id=file_id),
            )
            requests_.append(req)
        return self.batch_operate(requests_, resource="file")

    @assert_ok
    @handle_error
    def check_available(self, file_ids: str):
        """Check whether file_ids are available"""

        url = PcsNode.Available.url()
        data = dict(
            drive_id=self.default_drive_id,
            file_id_list=list(file_ids),
        )
        resp = self._request(Method.Post, url, json=data)
        return resp.json()

    @assert_ok
    @handle_error
    def share(self, *file_ids: str, password: str = "", period: int = 0, description: str = ""):
        """Share `remotepaths` to public

        period (int): The days for expiring. `0` means no expiring
        """

        expiration = ""  # Living forever
        if period > 0:
            expiration = timestamp_to_iso_8601(now_timestamp() + period * 24 * 60 * 60)

        url = PcsNode.Share.url()
        data = dict(
            drive_id=self.default_drive_id,
            expiration=expiration,
            file_id_list=list(file_ids),
            share_pwd=password,
            description=description,
        )
        resp = self._request(Method.Post, url, json=data)
        return resp.json()

    @assert_ok
    @handle_error
    def list_shared(self, next_marker: str = ""):
        """List shared links"""

        url = PcsNode.SharedList.url()
        data = dict(
            creator=self.user_id,
            include_canceled=False,
            order_by="created_at",
            order_direction="DESC",
            marker=next_marker,
        )
        resp = self._request(Method.Post, url, json=data)
        return resp.json()

    def cancel_shared(self, *share_ids: str):
        requests_ = []
        for share_id in share_ids:
            req = dict(
                method="POST",
                url="/share_link/cancel",
                id=share_id,
                headers={"Content-Type": "application/json"},
                body=dict(share_id=share_id),
            )
            requests_.append(req)

        return self.batch_operate(requests_, resource="file")

    @assert_ok
    @handle_error
    def get_share_token(self, share_id: str, share_password: str = ""):
        """Get share token"""

        shared_auth = self.__class__.SHARE_AUTHS.get(share_id)
        if shared_auth and not shared_auth.is_expired():
            share_password = share_password or shared_auth.share_password
            return shared_auth.info

        url = PcsNode.ShareToken.url()
        data = dict(share_id=share_id, share_pwd=share_password)
        resp = self._request(Method.Post, url, json=data)

        info = resp.json()
        if info.get("share_token"):
            # Store share password for refreshing share token
            self.__class__.SHARE_AUTHS[share_id] = SharedAuth.from_(share_id, share_password, info)

        return info

    def share_token(self, share_id: str) -> str:
        self.get_share_token(share_id)

        shared_auth = self.__class__.SHARE_AUTHS[share_id]
        return shared_auth.share_token

    @assert_ok
    @handle_error
    def shared_info(self, share_id: str):
        """Get shared items info"""

        url = PcsNode.SharedInfo.url()
        params = dict(share_id=share_id)
        data = dict(share_id=share_id)
        resp = self._request(Method.Post, url, params=params, json=data)
        info = resp.json()
        info["share_id"] = share_id
        return info

    # list_shared_files is an alias of list
    list_shared_files = list

    def transfer_shared_files(
        self, shared_file_ids: List[str], dest_id: str, share_id: str, auto_rename: bool = False
    ):
        """Transfer shared files to destination directory"""

        requests_ = []
        for file_id in shared_file_ids:
            req = dict(
                method="POST",
                url="/file/copy",
                id="0",
                headers={"Content-Type": "application/json"},
                body=dict(
                    auto_rename=auto_rename,
                    file_id=file_id,
                    share_id=share_id,
                    to_drive_id=self.default_drive_id,
                    to_parent_file_id=dest_id,
                ),
            )
            requests_.append(req)

        share_token = self.share_token(share_id)
        assert share_token, "Need share_token"

        headers = dict(PCS_HEADERS)
        headers["x-share-token"] = share_token

        return self.batch_operate(requests_, resource="file", headers=headers)

    @assert_ok
    @handle_error
    def _get_shared_file_download_url(self, shared_file_id: str, share_id: str, expire_duration: int = 10 * 60):
        url = PcsNode.SharedFileDownloadUrl.url()
        data = dict(expire_sec=expire_duration, file_id=shared_file_id, share_id=share_id)

        share_token = self.share_token(share_id)
        assert share_token, "Need share_token"

        headers = dict(PCS_HEADERS)
        headers["x-share-token"] = share_token

        resp = self._request(Method.Post, url, headers=headers, json=data)
        return resp.json()

    def shared_file_download_url(self, shared_file_id: str, share_id: str, expire_duration: int = 10 * 60) -> str:
        info = self._get_shared_file_download_url(shared_file_id, share_id, expire_duration=expire_duration)
        url = info["url"]

        headers = dict(PCS_HEADERS)
        headers["Referer"] = "https://www.aliyundrive.com/"
        resp = requests.get(url, headers=headers, allow_redirects=False)
        return resp.headers["Location"]

    @assert_ok
    @handle_error
    def user_info(self):
        url = PcsNode.User.url()
        resp = self._request(Method.Post, url, json={})
        info1 = resp.json()

        headers = dict(PCS_HEADERS)
        headers.update({"x-device-id": self.device_id, "x-signature": self.signature})
        url = PcsNode.PersonalInfo.url()
        resp = self._request(Method.Post, url, headers=headers, json={})
        info2 = resp.json()

        url = PcsNode.UserVip.url()
        resp = self._request(Method.Post, url, headers=headers, json={})
        info3 = resp.json()

        return {**info1, **info2, "user_vip_info": info3}

    @timeout_cache(1 * 60 * 60)  # 1 hour timeout
    @assert_ok
    @handle_error
    def download_link(self, file_id: str):
        url = PcsNode.DownloadUrl.url()
        data = dict(drive_id=self.default_drive_id, file_id=file_id)
        headers = dict(PCS_HEADERS)
        headers.update({"x-device-id": self.device_id, "x-signature": self.signature})
        resp = self._request(Method.Post, url, headers=headers, json=data)
        return resp.json()

    def file_stream(
        self,
        file_id: str,
        max_chunk_size: int = DEFAULT_MAX_CHUNK_SIZE,
        callback: Callable[..., None] = None,
        encrypt_password: bytes = b"",
    ) -> Optional[RangeRequestIO]:
        info = self.download_link(file_id)
        url = info["url"]

        headers = {
            "User-Agent": PCS_UA,
            "Connection": "Keep-Alive",
            "Referer": "https://www.aliyundrive.com/",
            "Accept-Encoding": "gzip, deflate, br",
            "Accept-Language": "en-US,en;q=0.9,zh-CN;q=0.8,zh;q=0.7,zh-TW;q=0.6",
        }
        return RangeRequestIO(
            Method.Get.value,
            url,
            headers=headers,
            max_chunk_size=max_chunk_size,
            callback=callback,
            encrypt_password=encrypt_password,
        )

    def shared_file_stream(
        self,
        shared_file_id: str,
        share_id: str,
        expire_duration: int = 10 * 60,
        max_chunk_size: int = DEFAULT_MAX_CHUNK_SIZE,
        callback: Callable[..., None] = None,
        encrypt_password: bytes = b"",
    ) -> Optional[RangeRequestIO]:
        url = self.shared_file_download_url(shared_file_id, share_id, expire_duration=expire_duration)

        headers = {
            "User-Agent": PCS_UA,
            "Connection": "Keep-Alive",
            "Referer": "https://www.aliyundrive.com/",
            "Accept-Encoding": "gzip, deflate, br",
            "Accept-Language": "en-US,en;q=0.9,zh-CN;q=0.8,zh;q=0.7,zh-TW;q=0.6",
        }
        return RangeRequestIO(
            Method.Get.value,
            url,
            headers=headers,
            max_chunk_size=max_chunk_size,
            callback=callback,
            encrypt_password=encrypt_password,
        )


class OpenPcsNode(Enum):
    f"""PCS Nodes which use {ALIYUNDRIVE_OPENAPI_DOMAIN}"""

    UpdateRefreshToken = "oauth/access_token"
    QrcodeLink = "oauth/authorize/qrcode"
    QrcodeUrl = "o/oauth/authorize?sid={sid}"
    QrcodeStatus = "oauth/qrcode/{sid}/status"

    Token = "v2/account/token"
    CreateSession = "users/v1/users/device/create_session"
    DriveInfo = "adrive/v1.0/user/getDriveInfo"

    Meta = "adrive/v1.0/openFile/get"
    FileList = "adrive/v1.0/openFile/list"
    Search = "adrive/v3/file/search"
    DownloadUrl = "adrive/v1.0/openFile/getDownloadUrl"

    CreateFile = "v2/file/create"
    UploadComplete = "v2/file/complete"

    Batch = "v3/batch"

    CreateWithFolders = "adrive/v2/file/createWithFolders"
    Move = "adrive/v3/file/move"
    Update = "v3/file/update"
    Trash = "v2/recyclebin/trash"

    Available = "adrive/v2/share_link/check_avaliable"
    Share = "adrive/v2/share_link/create"
    ShareToken = "v2/share_link/get_share_token"
    SharedInfo = "adrive/v3/share_link/get_share_by_anonymous"
    SharedList = "adrive/v3/share_link/list"
    SharedFileDownloadUrl = "v2/file/get_share_link_download_url"

    SpaceInfo = "adrive/v1.0/user/getSpaceInfo"
    # PersonalInfo = "v2/databox/get_personal_info"
    PersonalInfo = "adrive/v1.0/user/get"
    User = "v2/user/get"
    UserVip = "adrive/v1.0/user/getVipInfo"

    def url(self) -> str:
        return f"{ALIYUNDRIVE_OPENAPI_DOMAIN}/{self.value}"


class AliOpenPCS:
    """`Aliyundrive Open PCS` provides pcs's apis which return raw json"""

    SHARE_AUTHS: Dict[str, SharedAuth] = {}

    def __init__(
        self,
        refresh_token: str,
        access_token: str = "",
        token_type: str = "Bearer",
        expire_time: int = 0,
        client_id: str = "",
        client_secret: str = "",
        client_server: str = "",
        user_id: str = "",
        user_name: str = "",
        nick_name: str = "",
        default_drive_id: str = "",
        role: str = "",
        status: str = "",
    ):
        assert (
            client_id and client_secret
        ) or client_server, "(client_id and client_secret) or client_server must be set"

        self._session = requests.Session()

        self._refresh_token = refresh_token
        self._access_token = access_token
        self._token_type = token_type
        self._expire_time = expire_time

        self._client_id = client_id
        self._client_secret = client_secret
        self._client_server = client_server

        self._user_id = user_id
        self._user_name = user_name
        self._nick_name = nick_name
        self._default_drive_id = default_drive_id
        self._role = role
        self._status = status
        self._nonce = 0

    def __str__(self) -> str:
        return f"""AliPCS

    user_id: {self.user_id}
    user_name: {self.user_name}
    nick_name: {self.nick_name}

    refresh_token: {self.refresh_token}
    access_token: {self.access_token}
    token_type: {self.token_type}
    expire_time: {self.expire_time}

    client_id: {self._client_id}
    client_secret: {self._client_secret}
    client_server: {self._client_server}

    default_drive_id: {self.default_drive_id}
    role: {self.role}
    status: {self.status}
    nonce: {self._nonce}
    """

    @property
    def refresh_token(self) -> str:
        now = now_timestamp()
        if self._expire_time - now < 200 or not self._access_token:
            self._update_refresh_token()

        if not self._default_drive_id:
            self._get_drive_info()

        return self._refresh_token

    @property
    def access_token(self) -> str:
        self.refresh_token
        return self._access_token

    @property
    def client_id(self) -> str:
        return self._client_id

    @property
    def client_secret(self) -> str:
        return self._client_secret

    @property
    def client_server(self) -> str:
        return self._client_server

    @property
    def user_id(self) -> str:
        self.refresh_token
        return self._user_id

    @property
    def user_name(self) -> str:
        self.refresh_token
        return self._user_name

    @property
    def nick_name(self) -> str:
        self.refresh_token
        return self._nick_name

    @property
    def token_type(self) -> str:
        self.refresh_token
        return self._token_type

    @property
    def expire_time(self) -> int:
        self.refresh_token
        return self._expire_time

    @property
    def default_drive_id(self) -> str:
        self.refresh_token
        return self._default_drive_id

    @property
    def role(self) -> str:
        self.refresh_token
        return self._role

    @property
    def status(self) -> str:
        self.refresh_token
        return self._status

    def _request(
        self,
        method: Method,
        url: str,
        params: Optional[Dict[str, str]] = None,
        headers: Optional[Dict[str, str]] = None,
        data: Union[str, bytes, Dict[str, str], Any] = None,
        json: Any = None,
        files: Optional[Dict[str, Any]] = None,
        refresh: bool = False,
        **kwargs,
    ) -> requests.Response:
        if not headers:
            headers = dict(PCS_HEADERS)

        if not refresh and "Authorization" not in headers:
            headers["Authorization"] = f"{self._token_type} {self._access_token}"

        if json is not None:
            headers["Content-Type"] = "application/json;charset=UTF-8"

        if isinstance(data, (MultipartEncoder, MultipartEncoderMonitor)):
            headers["Content-Type"] = data.content_type

        try:
            resp = self._session.request(
                method.value, url, params=params, headers=headers, data=data, json=json, files=files, **kwargs
            )
            return resp
        except Exception as err:
            raise AliPCSError("AliPCS._request", cause=err)

    def _update_refresh_token(self):
        """Update refresh token"""

        data = dict(refresh_token=self._refresh_token, grant_type="refresh_token")
        if self._client_id and self._client_secret:
            url = OpenPcsNode.UpdateRefreshToken.url()
            data["client_id"] = self._client_id
            data["client_secret"] = self._client_secret
        else:
            url = urljoin(self._client_server, OpenPcsNode.UpdateRefreshToken.value)

        resp = self._request(Method.Post, url, json=data, refresh=True)
        info = resp.json()

        if "code" in info:
            err = parse_error(info["code"], info=info)
            raise err

        self._refresh_token = info["refresh_token"]
        self._access_token = info["access_token"]
        self._token_type = info["token_type"]
        self._expire_time = now_timestamp() + info["expires_in"]

        return info

    def _get_drive_info(self):
        """Get drive info"""

        url = OpenPcsNode.DriveInfo.url()
        data = dict()

        resp = self._request(Method.Post, url, json=data)
        info = resp.json()

        if "code" in info:
            err = parse_error(info["code"], info=info)
            raise err

        self._user_id = info["user_id"]
        self._user_name = info["user_name"]
        self._nick_name = info["nick_name"]
        self._default_drive_id = info["default_drive_id"]
        self._domain_id = info["domain_id"]
        self._role = info["role"]
        self._status = info["status"]

        return info

    def meta(self, *file_ids: str, share_id: str = None):
        assert "root" not in file_ids, '"root" has NOT meta info'

        responses = []
        for file_id in file_ids:
            data = dict(file_id=file_id, drive_id=self.default_drive_id)
            url = PcsNode.Meta.url()
            resp = self._request(Method.Post, url, json=data)
            info = resp.json()
            responses.append(dict(body=info))

        return dict(responses=responses)

    def exists(self, file_id: str) -> bool:
        if file_id == "root":
            return True

        r = self.meta(file_id)
        info = r["responses"][0]["body"]
        if info.get("code") == "NotFound.File":
            return False
        else:
            return True

    def is_file(self, file_id: str) -> bool:
        if file_id == "root":
            return False

        r = self.meta(file_id)
        info = r["responses"][0]["body"]
        if info.get("code") == "NotFound.File":
            return False
        if info["type"] == "file":
            return True
        else:
            return False

    def is_dir(self, file_id: str) -> bool:
        if file_id == "root":
            return True

        r = self.meta(file_id)
        info = r["responses"][0]["body"]
        if info.get("code") == "NotFound.File":
            return False
        if info["type"] == "folder":
            return True
        else:
            return False

    @assert_ok
    @handle_error
    def list(
        self,
        file_id: str,
        share_id: str = None,
        desc: bool = False,
        name: bool = False,
        time: bool = False,
        size: bool = False,
        all: bool = False,
        limit: int = 200,
        url_expire_sec: int = 14200,
        next_marker: str = "",
    ):
        """List files at the directory which has `file_id`

        Args:
            file_id (str):
                The file_id of the directory
        """

        assert limit <= 200, "`limit` should be less than 200"

        url = PcsNode.FileList.url()
        orderby = "name"
        if name:
            orderby = "name"
        elif time:
            orderby = "updated_at"  # "created_at"  # 服务器最后修改时间
        elif size:
            orderby = "size"

        data = dict(
            all=all,
            drive_id=self.default_drive_id,
            fields="*",
            limit=limit,
            order_by=orderby,
            order_direction="DESC" if desc else "ASC",
            parent_file_id=file_id,
            url_expire_sec=url_expire_sec,
            marker=next_marker,
        )

        headers = dict(PCS_HEADERS)
        resp = self._request(Method.Post, url, headers=headers, json=data)
        return resp.json()

    @staticmethod
    def part_info_list(part_number: int) -> List[Dict[str, int]]:
        return [dict(part_number=i) for i in range(1, part_number + 1)]

    @timeout_cache(4 * 60 * 60)  # 4 hour timeout
    @assert_ok
    @handle_error
    def download_link(self, file_id: str):
        url = OpenPcsNode.DownloadUrl.url()
        data = dict(drive_id=self.default_drive_id, file_id=file_id, expire_sec=14400)
        resp = self._request(Method.Post, url, json=data)
        return resp.json()

    def file_stream(
        self,
        file_id: str,
        max_chunk_size: int = DEFAULT_MAX_CHUNK_SIZE,
        callback: Callable[..., None] = None,
        encrypt_password: bytes = b"",
    ) -> Optional[RangeRequestIO]:
        info = self.download_link(file_id)
        url = info["url"]

        headers = {
            "User-Agent": PCS_UA,
            "Connection": "Keep-Alive",
            "Referer": "https://www.aliyundrive.com/",
            "Accept-Encoding": "gzip, deflate, br",
            "Accept-Language": "en-US,en;q=0.9,zh-CN;q=0.8,zh;q=0.7,zh-TW;q=0.6",
        }
        return RangeRequestIO(
            Method.Get.value,
            url,
            headers=headers,
            max_chunk_size=max_chunk_size,
            callback=callback,
            encrypt_password=encrypt_password,
        )


class AliOpenAuth:
    def __init__(self, client_id: str = "", client_secret: str = "", client_server: str = ""):
        assert (
            client_id and client_secret
        ) or client_server, "(client_id and client_secret) or client_server must be set"

        self._session = requests.Session()

        self._client_id = client_id
        self._client_secret = client_secret
        self._client_server = client_server

    def _request(
        self,
        method: Method,
        url: str,
        params: Optional[Dict[str, str]] = None,
        headers: Optional[Dict[str, str]] = None,
        data: Union[str, bytes, Dict[str, str], Any] = None,
        json: Any = None,
        files: Optional[Dict[str, Any]] = None,
        **kwargs,
    ) -> requests.Response:
        if not headers:
            headers = dict(PCS_HEADERS)

        try:
            resp = self._session.request(
                method.value, url, params=params, headers=headers, data=data, json=json, files=files, **kwargs
            )
            return resp
        except Exception as err:
            raise AliPCSError("AliOpenAuth._request", cause=err)

    @staticmethod
    def qrcode_url(sid: str) -> str:
        return OpenPcsNode.QrcodeUrl.url().format(sid)

    @handle_error
    def get_qrcode_info(self):
        data: Dict[str, Any] = dict(
            scopes=[
                "user:base",
                "file:all:read",
                "file:all:write",
            ],
            width=None,
            height=None,
        )

        if self._client_server:
            url = self._client_server.strip("/") + "/" + OpenPcsNode.QrcodeLink.value
        else:
            data["client_id"] = self._client_id
            data["client_secret"] = self._client_secret
            url = OpenPcsNode.QrcodeLink.url()

        resp = self._request(Method.Post, url, json=data)
        return resp.json()

    @handle_error
    def scan_status(self, sid: str):
        url = OpenPcsNode.QrcodeStatus.url().format(sid=sid)
        resp = self._request(Method.Get, url)
        return resp.json()

    @handle_error
    def get_refresh_token(self, auth_code: str):
        data = dict(grant_type="authorization_code", code=auth_code)

        if self._client_server:
            url = self._client_server.strip("/") + "/" + OpenPcsNode.UpdateRefreshToken.value
        else:
            data["client_id"] = self._client_id
            data["client_secret"] = self._client_secret
            url = OpenPcsNode.UpdateRefreshToken.url()

        resp = self._request(Method.Post, url, json=data)
        return resp.json()



from typing import Tuple, Union, Iterator
from pathlib import Path, PurePosixPath
import os
from os import PathLike

from alipcs_py.common.platform import IS_WIN

PathType = Union["str", PathLike, Path]


def exists(localpath: PathType) -> bool:
    localpath = Path(localpath)
    return localpath.exists()


def is_file(localpath: PathType) -> bool:
    localpath = Path(localpath)
    return localpath.is_file()


def is_dir(localpath: PathType) -> bool:
    localpath = Path(localpath)
    return localpath.is_dir()


def walk(localpath: PathType) -> Iterator[str]:
    for root, _, files in os.walk(Path(localpath)):
        r = Path(root)
        for fl in files:
            yield (r / fl).as_posix()


def join_path(parent: PathType, *children: PathType) -> str:
    """Join posix paths"""

    _path = Path(parent)
    for child in children:
        _path = _path / child

    path = _path.as_posix()
    has_root = path.startswith("/")
    if not has_root:
        path = "/" + path

    path = Path(path).resolve().as_posix()

    if IS_WIN:
        p = path.split(":", 1)[-1]
        if not has_root:
            return p[1:]
        else:
            return p
    else:
        if not has_root:
            return path[1:]
        else:
            return path


def split_posix_path(path: PathType) -> Tuple[str, ...]:
    return PurePosixPath(path).parts


def posix_path_basename(path: PathType) -> str:
    return PurePosixPath(path).name


def posix_path_dirname(path: PathType) -> str:
    return PurePosixPath(path).parent.as_posix()



from typing import Dict, Any
import json


def json_dumps(obj: Any) -> str:
    return json.dumps(obj, separators=(",", ":"), ensure_ascii=False, sort_keys=True)


def json_dump_values(obj: Dict[str, Any]) -> Dict[str, str]:
    """Convert dict's values which are list or dict as json string"""

    new_obj = dict(obj)
    for k, v in new_obj.items():
        if isinstance(v, (dict, list)):
            new_obj[k] = json.dumps(v)
    return new_obj



from typing import Optional, List

from alipcs_py.alipcs import AliPCSApi
from alipcs_py.alipcs.inner import PcsFile
from alipcs_py.common.path import join_path
from alipcs_py.commands.log import get_logger
from alipcs_py.commands.sifter import Sifter, sift
from alipcs_py.commands.display import display_files

from rich import print

logger = get_logger(__name__)

DEFAULT_MAX_WORKERS = 10


def list_file(
    api: AliPCSApi,
    remotepath: str,
    file_id: str = None,
    share_id: str = None,
    desc: bool = False,
    name: bool = False,
    time: bool = False,
    size: bool = False,
    all: bool = False,
    limit: int = 200,
    url_expire_sec: int = 14400,
    recursive: bool = False,
    sifters: List[Sifter] = [],
    highlight: bool = False,
    rapiduploadinfo_file: Optional[str] = None,
    user_id: Optional[str] = None,
    user_name: Optional[str] = None,
    show_size: bool = False,
    show_date: bool = False,
    show_file_id: bool = False,
    show_hash: bool = False,
    show_absolute_path: bool = False,
    show_dl_link: bool = False,
    csv: bool = False,
    only_dl_link: bool = False,
):
    pcs_file: Optional[PcsFile]
    if file_id:
        pcs_file = api.meta(file_id, share_id=share_id)[0]
    else:
        pcs_file = api.path(remotepath, share_id=share_id)
    if not pcs_file:
        return

    is_dir = pcs_file.is_dir
    if is_dir:
        pcs_files = api.list_path(
            remotepath,
            file_id=pcs_file.file_id,
            share_id=share_id,
            desc=desc,
            name=name,
            time=time,
            size=size,
            all=all,
            limit=limit,
            url_expire_sec=url_expire_sec,
        )
    else:
        pcs_files = [pcs_file]

    pcs_files = sift(pcs_files, sifters, recursive=recursive)
    if not pcs_files:
        return

    if show_dl_link:
        for pcs_file in pcs_files:
            if only_dl_link:
                print(pcs_file.download_url)

    if not only_dl_link:
        display_files(
            pcs_files,
            remotepath,
            sifters=sifters,
            highlight=highlight,
            show_size=show_size,
            show_date=show_date,
            show_file_id=show_file_id,
            show_hash=show_hash,
            show_absolute_path=show_absolute_path,
            show_dl_link=show_dl_link,
            csv=csv,
        )

    if is_dir and recursive:
        for pcs_file in pcs_files:
            if pcs_file.is_dir:
                list_file(
                    api,
                    join_path(remotepath, pcs_file.name),
                    file_id=pcs_file.file_id,
                    share_id=share_id,
                    desc=desc,
                    name=name,
                    time=time,
                    size=size,
                    all=all,
                    limit=limit,
                    url_expire_sec=url_expire_sec,
                    recursive=recursive,
                    sifters=sifters,
                    highlight=highlight,
                    rapiduploadinfo_file=rapiduploadinfo_file,
                    user_id=user_id,
                    user_name=user_name,
                    show_size=show_size,
                    show_date=show_date,
                    show_file_id=show_file_id,
                    show_hash=show_hash,
                    show_absolute_path=show_absolute_path,
                    show_dl_link=show_dl_link,
                    csv=csv,
                    only_dl_link=only_dl_link,
                )


def list_files(
    api: AliPCSApi,
    *remotepaths: str,
    file_ids: List[str] = [],
    share_id: str = None,
    desc: bool = False,
    name: bool = False,
    time: bool = False,
    size: bool = False,
    all: bool = False,
    limit: int = 200,
    url_expire_sec: int = 14400,
    recursive: bool = False,
    sifters: List[Sifter] = [],
    highlight: bool = False,
    rapiduploadinfo_file: Optional[str] = None,
    user_id: Optional[str] = None,
    user_name: Optional[str] = None,
    show_size: bool = False,
    show_date: bool = False,
    show_file_id: bool = False,
    show_hash: bool = False,
    show_absolute_path: bool = False,
    show_dl_link: bool = False,
    csv: bool = False,
    only_dl_link: bool = False,
):
    for rp in remotepaths:
        list_file(
            api,
            rp,
            share_id=share_id,
            desc=desc,
            name=name,
            time=time,
            size=size,
            all=all,
            limit=limit,
            url_expire_sec=url_expire_sec,
            recursive=recursive,
            sifters=sifters,
            highlight=highlight,
            rapiduploadinfo_file=rapiduploadinfo_file,
            user_id=user_id,
            user_name=user_name,
            show_size=show_size,
            show_date=show_date,
            show_file_id=show_file_id,
            show_hash=show_hash,
            show_absolute_path=show_absolute_path,
            show_dl_link=show_dl_link,
            csv=csv,
            only_dl_link=only_dl_link,
        )

    for file_id in file_ids:
        list_file(
            api,
            "",
            file_id=file_id,
            share_id=share_id,
            desc=desc,
            name=name,
            time=time,
            size=size,
            all=all,
            limit=limit,
            url_expire_sec=url_expire_sec,
            recursive=recursive,
            sifters=sifters,
            highlight=highlight,
            rapiduploadinfo_file=rapiduploadinfo_file,
            user_id=user_id,
            user_name=user_name,
            show_size=show_size,
            show_date=show_date,
            show_file_id=show_file_id,
            show_hash=show_hash,
            show_absolute_path=show_absolute_path,
            show_dl_link=show_dl_link,
            csv=csv,
            only_dl_link=only_dl_link,
        )
